# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2018-2020 Andrew Rechnitzer
# Copyright (C) 2019-2020 Colin B. Macdonald

"""
Backend bits n bobs to talk to the server
"""

__author__ = "Andrew Rechnitzer, Colin B. Macdonald"
__copyright__ = "Copyright (C) 2018-2020 Andrew Rechnitzer, Colin B. Macdonald"
__credits__ = ["Andrew Rechnitzer", "Colin Macdonald", "Elvis Cai", "Matt Coles"]
__license__ = "AGPL-3.0-or-later"


import json
import ssl
import threading
import hashlib
import logging
from io import StringIO, BytesIO

import urllib3
import requests
from requests_toolbelt import MultipartEncoder, MultipartDecoder

from plom import __version__, Plom_API_Version, Default_Port
from plom.plom_exceptions import PlomBenignException, PlomSeriousException
from plom.plom_exceptions import (
    PlomAuthenticationException,
    PlomAPIException,
    PlomExistingLoginException,
    PlomConflict,
    PlomTakenException,
    PlomNoMoreException,
    PlomRangeException,
    PlomLatexException,
    PlomTaskChangedError,
    PlomTaskDeletedError,
)

log = logging.getLogger("messenger")
# requests_log = logging.getLogger("urllib3")
# requests_log.setLevel(logging.DEBUG)
# requests_log.propagate = True

# If we use unverified ssl certificates we get lots of warnings,
# so put in this to hide them.
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class BaseMessenger:
    """Basic communication with a Plom Server.

    Handles authentication and other common tasks; subclasses can add
    other features.
    """

    def __init__(self, s=None, port=Default_Port):
        sslContext = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH)
        sslContext.check_hostname = False
        # Server defaults
        self.session = None
        self.user = None
        self.token = None
        if s:
            server = s
        else:
            server = "127.0.0.1"
        self.server = "{}:{}".format(server, port)
        message_port = None  # REMOVE
        self.SRmutex = threading.Lock()
        # base = "https://{}:{}/".format(s, mp)

    def whoami(self):
        return self.user

    def start(self):
        """Start the messenger session"""
        if self.session:
            log.debug("already have an requests-session")
        else:
            log.debug("starting a new requests-session")
            self.session = requests.Session()
            # TODO - UBC wifi is crappy: have some sort of "hey you've retried
            # nn times already, are you sure you want to keep retrying" message.
            self.session.mount("https://", requests.adapters.HTTPAdapter(max_retries=3))
        try:
            response = self.session.get(
                "https://{}/Version".format(self.server),
                verify=False,
            )
            response.raise_for_status()
        except requests.ConnectionError as err:
            raise PlomBenignException(
                "Cannot connect to server. Please check server details."
            ) from None
        r = response.text
        return r

    def stop(self):
        """Stop the messenger"""
        if self.session:
            log.debug("stopping requests-session")
            self.session.close()
            self.session = None

    def isStarted(self):
        return bool(self.session)

    # ------------------------
    # ------------------------
    # Authentication stuff

    def requestAndSaveToken(self, user, pw):
        """Get a authorisation token from the server.

        The token is then used to authenticate future transactions with the server.

        """
        self.SRmutex.acquire()
        try:
            response = self.session.put(
                "https://{}/users/{}".format(self.server, user),
                json={
                    "user": user,
                    "pw": pw,
                    "api": Plom_API_Version,
                    "client_ver": __version__,
                },
                verify=False,
                timeout=5,
            )
            # throw errors when response code != 200.
            response.raise_for_status()
            # convert the content of the response to a textfile for identifier
            self.token = response.json()
            self.user = user
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException(response.json()) from None
            elif response.status_code == 400:  # API error
                raise PlomAPIException(response.json()) from None
            elif response.status_code == 409:
                raise PlomExistingLoginException(response.json()) from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        except requests.ConnectionError as err:
            raise PlomSeriousException(
                "Cannot connect to server\n {}\n Please check details before trying again.".format(
                    self.server
                )
            ) from None
        finally:
            self.SRmutex.release()

    def clearAuthorisation(self, user, pw):
        self.SRmutex.acquire()
        try:
            response = self.session.delete(
                "https://{}/authorisation".format(self.server),
                json={"user": user, "password": pw},
                verify=False,
            )
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

    def closeUser(self):
        self.SRmutex.acquire()
        try:
            response = self.session.delete(
                "https://{}/users/{}".format(self.server, self.user),
                json={"user": self.user, "token": self.token},
                verify=False,
            )
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        return True

    # ----------------------
    # ----------------------
    # Test information

    def getInfoShortName(self):
        self.SRmutex.acquire()
        try:
            response = self.session.get(
                "https://{}/info/shortName".format(self.server), verify=False
            )
            response.raise_for_status()
            shortName = response.text
        except requests.HTTPError as e:
            if response.status_code == 404:
                raise PlomSeriousException(
                    "Server could not find the spec - this should not happen!"
                ) from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        return shortName

    def get_spec(self):
        """Get the specification of the exam from the server.

        Returns:
            dict: the server's spec file, as in :func:`plom.SpecVerifier`.
        """
        self.SRmutex.acquire()
        try:
            response = self.session.get(
                "https://{}/info/spec".format(self.server),
                verify=False,
            )
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 404:
                raise PlomSeriousException("Server could not find the spec") from None
            else:
                raise PlomSeriousException("Some other sort of error {}".format(e))
        finally:
            self.SRmutex.release()

        return response.json()

    def getInfoGeneral(self):
        """Get some info from pre-0.5.0 server which don't expose the spec.

        Probably we can deprecate or remove this.  Old clients trying to
        talk to newer servers will just get a 404.

        Returns:
            dict: some of the fields of the server's spec file.
        """
        self.SRmutex.acquire()
        try:
            response = self.session.get(
                "https://{}/info/general".format(self.server),
                verify=False,
            )
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 404:
                raise PlomSeriousException("Server could not find the spec") from None
            else:
                raise PlomSeriousException("Some other sort of error {}".format(e))
        finally:
            self.SRmutex.release()

        pv = response.json()
        fields = (
            "name",
            "numberToProduce",
            "numberOfPages",
            "numberOfQuestions",
            "numberOfVersions",
            "publicCode",
        )
        return dict(zip(fields, pv))

    def IDrequestClasslist(self):
        """Ask server for the classlist.

        Returns:
            list: ordered list of (student id, student name) pairs.
                Both are strings.

        Raises:
            PlomAuthenticationException: login troubles.
            PlomBenignException: server has no classlist.
            PlomSeriousException: all other failures.
        """
        self.SRmutex.acquire()
        try:
            response = self.session.get(
                "https://{}/ID/classlist".format(self.server),
                json={"user": self.user, "token": self.token},
                verify=False,
            )
            # throw errors when response code != 200.
            response.raise_for_status()
            # you can assign to the encoding to override the autodetection
            # TODO: define API such that classlist must be utf-8?
            # print(response.encoding)
            # response.encoding = 'utf-8'
            # classlist = StringIO(response.text)
            classlist = response.json()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            elif response.status_code == 404:
                raise PlomBenignException("Server cannot find the class list") from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        return classlist


class Messenger(BaseMessenger):
    """Handle communication with a Plom Server."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # ------------------------
    # ------------------------
    # ID client API stuff

    def IDprogressCount(self):
        self.SRmutex.acquire()
        try:
            response = self.session.get(
                "https://{}/ID/progress".format(self.server),
                json={"user": self.user, "token": self.token},
                verify=False,
            )
            # throw errors when response code != 200.
            response.raise_for_status()
            # convert the content of the response to a textfile for identifier
            progress = response.json()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        return progress

    def IDaskNextTask(self):
        """Return the TGV of a paper that needs IDing.

        Return:
            string or None if no papers need IDing.

        Raises:
            SeriousError: if something has unexpectedly gone wrong.
        """
        self.SRmutex.acquire()
        try:
            response = self.session.get(
                "https://{}/ID/tasks/available".format(self.server),
                json={"user": self.user, "token": self.token},
                verify=False,
            )
            # throw errors when response code != 200.
            response.raise_for_status()
            if response.status_code == 204:
                return None
            tgv = response.json()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        return tgv

    def IDrequestPredictions(self):
        self.SRmutex.acquire()
        try:
            response = self.session.get(
                "https://{}/ID/predictions".format(self.server),
                json={"user": self.user, "token": self.token},
                verify=False,
            )
            response.raise_for_status()
            # TODO: print(response.encoding) autodetected
            predictions = StringIO(response.text)
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            elif response.status_code == 404:
                raise PlomSeriousException(
                    "Server cannot find the prediction list."
                ) from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        return predictions

    def IDrequestDoneTasks(self):
        self.SRmutex.acquire()
        try:
            response = self.session.get(
                "https://{}/ID/tasks/complete".format(self.server),
                json={"user": self.user, "token": self.token},
                verify=False,
            )
            response.raise_for_status()
            idList = response.json()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        return idList

    def IDrequestImage(self, code):
        self.SRmutex.acquire()
        try:
            response = self.session.get(
                "https://{}/ID/images/{}".format(self.server, code),
                json={"user": self.user, "token": self.token},
                verify=False,
            )
            response.raise_for_status()
            imageList = []
            for img in MultipartDecoder.from_response(response).parts:
                imageList.append(
                    BytesIO(img.content).getvalue()
                )  # pass back image as bytes
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            elif response.status_code == 404:
                raise PlomSeriousException(
                    "Cannot find image file for {}.".format(code)
                ) from None
            elif response.status_code == 409:
                raise PlomSeriousException(
                    "Another user has the image for {}. This should not happen".format(
                        code
                    )
                ) from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        return imageList

    # ------------------------

    # TODO - API needs improve. Both of these throw a put/patch to same url = /ID/tasks/{tgv}
    # One only updates the user claim, while the other actually ID's it.
    # Think of better url structure for this?
    def IDclaimThisTask(self, code):
        self.SRmutex.acquire()
        try:
            response = self.session.patch(
                "https://{}/ID/tasks/{}".format(self.server, code),
                json={"user": self.user, "token": self.token},
                verify=False,
            )
            if response.status_code == 204:
                raise PlomTakenException("Task taken by another user.")
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        imageList = []
        for img in MultipartDecoder.from_response(response).parts:
            imageList.append(
                BytesIO(img.content).getvalue()
            )  # pass back image as bytes
        return imageList

    def IDreturnIDdTask(self, code, studentID, studentName):
        """Return a completed IDing task: identify a paper.

        Exceptions:
            PlomConflict: `studentID` already used on a different paper.
            PlomAuthenticationException: login problems.
            PlomSeriousException: other errors.
        """
        self.SRmutex.acquire()
        try:
            response = self.session.put(
                "https://{}/ID/tasks/{}".format(self.server, code),
                json={
                    "user": self.user,
                    "token": self.token,
                    "sid": studentID,
                    "sname": studentName,
                },
                verify=False,
            )
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 409:
                raise PlomConflict(e) from None
            elif response.status_code == 401:
                raise PlomAuthenticationException() from None
            elif response.status_code == 404:
                raise PlomSeriousException(e) from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        # TODO - do we need this return value?
        return True

    def IDdidNotFinishTask(self, code):
        self.SRmutex.acquire()
        try:
            response = self.session.delete(
                "https://{}/ID/tasks/{}".format(self.server, code),
                json={"user": self.user, "token": self.token},
                verify=False,
            )
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        return True

    # ------------------------
    # ------------------------
    # Marker stuff
    def MgetMaxMark(self, question, ver):
        """Get the maximum mark for this question and version.

        Raises:
            PlomRangeExeception: `question` or `ver` is out of range.
            PlomAuthenticationException:
            PlomSeriousException: something unexpected happened.
        """
        self.SRmutex.acquire()
        try:
            response = self.session.get(
                "https://{}/MK/maxMark".format(self.server),
                json={"user": self.user, "token": self.token, "q": question, "v": ver},
                verify=False,
            )
            # throw errors when response code != 200.
            response.raise_for_status()
            # convert the content of the response to a textfile for identifier
            maxMark = response.json()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            elif response.status_code == 416:
                raise PlomRangeException(response.text) from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        return maxMark

    def MdidNotFinishTask(self, code):
        self.SRmutex.acquire()
        try:
            response = self.session.delete(
                "https://{}/MK/tasks/{}".format(self.server, code),
                json={"user": self.user, "token": self.token},
                verify=False,
            )
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        return True

    def MrequestDoneTasks(self, q, v):
        self.SRmutex.acquire()
        try:
            response = self.session.get(
                "https://{}/MK/tasks/complete".format(self.server),
                json={"user": self.user, "token": self.token, "q": q, "v": v},
                verify=False,
            )
            response.raise_for_status()
            mList = response.json()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        return mList

    def MprogressCount(self, q, v):
        self.SRmutex.acquire()
        try:
            response = self.session.get(
                "https://{}/MK/progress".format(self.server),
                json={"user": self.user, "token": self.token, "q": q, "v": v},
                verify=False,
            )
            # throw errors when response code != 200.
            response.raise_for_status()
            # convert the content of the response to a textfile for identifier
            progress = response.json()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        return progress

    def MaskNextTask(self, q, v):
        """Ask server for a new marking task, return tgv or None.

        None indicated no more tasks available.
        TODO: why are we using json for a string return?
        """

        self.SRmutex.acquire()
        try:
            response = self.session.get(
                "https://{}/MK/tasks/available".format(self.server),
                json={"user": self.user, "token": self.token, "q": q, "v": v},
                verify=False,
            )
            # throw errors when response code != 200.
            if response.status_code == 204:
                return None
            response.raise_for_status()
            # convert the content of the response to a textfile for identifier
            tgv = response.json()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        return tgv

    def MclaimThisTask(self, code):
        """Claim a task from server and get back metadata.

        args:
            code (str): a task code such as `"q0123g2"`.

        returns:
            list: Consisting of image_metadata, tags, integrity_check.
        """

        self.SRmutex.acquire()
        try:
            response = self.session.patch(
                "https://{}/MK/tasks/{}".format(self.server, code),
                json={"user": self.user, "token": self.token},
                verify=False,
            )
            response.raise_for_status()
            if response.status_code == 204:
                raise PlomTakenException("Task taken by another user.")
            ret = response.json()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()
        return ret

    def MlatexFragment(self, latex):
        self.SRmutex.acquire()
        try:
            response = self.session.get(
                "https://{}/MK/latex".format(self.server),
                json={"user": self.user, "token": self.token, "fragment": latex},
                verify=False,
            )
            response.raise_for_status()
            image = BytesIO(response.content).getvalue()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            elif response.status_code == 406:
                raise PlomLatexException(
                    "There is an error in your latex fragment"
                ) from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        return image

    def MrequestImages(self, code, integrity_check):
        """Download images relevant to a question, both original and annotated.

        Args:
            code (str): the task code such as "q1234g9".

        Returns:
            3-tuple: `(image_metadata, annotated_image, plom_file)`
                `image_metadata` has various stuff: DB ids, md5sums, etc
                `annotated_image` and `plom_file` are the png file and
                and data associated with a previous annotations, or None.

        Raises:
            PlomAuthenticationException
            PlomTaskChangedError: you no longer own this task.
            PlomTaskDeletedError
            PlomSeriousException
        """
        self.SRmutex.acquire()
        try:
            response = self.session.get(
                "https://{}/MK/images/{}".format(self.server, code),
                json={
                    "user": self.user,
                    "token": self.token,
                    "integrity_check": integrity_check,
                },
                verify=False,
            )
            response.raise_for_status()

            # response is either [metadata] or [metadata, annotated_image, plom_file]
            imagesAnnotAndPlom = MultipartDecoder.from_response(response).parts
            image_metadata = json.loads(imagesAnnotAndPlom[0].text)
            if len(imagesAnnotAndPlom) == 1:
                # all is fine - no annotated image or plom data
                anImage = None
                plDat = None
            elif len(imagesAnnotAndPlom) == 3:
                # all fine - last two parts are annotated image + plom-data
                anImage = BytesIO(imagesAnnotAndPlom[1].content).getvalue()
                plDat = BytesIO(imagesAnnotAndPlom[2].content).getvalue()
            else:
                raise PlomSeriousException(
                    "Number of returns doesn't make sense: should be 1 or 3 but is {}".format(
                        len(imagesAnnotAndPlom)
                    )
                )
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            elif response.status_code == 404:
                raise PlomSeriousException(
                    "Cannot find image file for {}.".format(code)
                ) from None
            elif response.status_code == 409:
                raise PlomTaskChangedError(
                    "Ownership of task {} has changed.".format(code)
                ) from None
            elif response.status_code == 406:
                raise PlomTaskChangedError(
                    "Task {} has been changed by manager.".format(code)
                ) from None
            elif response.status_code == 410:
                raise PlomTaskDeletedError(
                    "Task {} has been deleted by manager.".format(code)
                ) from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        return image_metadata, anImage, plDat

    def MrequestOriginalImages(self, task):
        self.SRmutex.acquire()
        try:
            response = self.session.get(
                "https://{}/MK/originalImages/{}".format(self.server, task),
                json={"user": self.user, "token": self.token},
                verify=False,
            )
            if response.status_code == 204:
                raise PlomNoMoreException("No task = {}.".format(task))
            response.raise_for_status()
            # response is [image1, image2,... image.n]
            imageList = []
            for img in MultipartDecoder.from_response(response).parts:
                imageList.append(BytesIO(img.content).getvalue())

        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            elif response.status_code == 404:
                raise PlomNoMoreException(
                    "Cannot find image file for {}".format(task)
                ) from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        return imageList

    def MreturnMarkedTask(
        self,
        code,
        pg,
        ver,
        score,
        mtime,
        tags,
        aname,
        pname,
        cname,
        integrity_check,
        image_md5_list,
    ):
        """Upload annotated image and associated data to the server.

        Returns:
            list: a 2-list of the form `[#done, #total]`.

        Raises:
            PlomAuthenticationException
            PlomConflict: integrity check failed, perhaps manager
                altered task.
            PlomTaskChangedError
            PlomTaskDeletedError
            PlomSeriousException
        """
        self.SRmutex.acquire()
        try:
            # doesn't like ints, so convert ints to strings
            param = {
                "user": self.user,
                "token": self.token,
                "pg": str(pg),
                "ver": str(ver),
                "score": str(score),
                "mtime": str(mtime),
                "tags": tags,
                "comments": open(cname, "r").read(),
                "md5sum": hashlib.md5(open(aname, "rb").read()).hexdigest(),
                "integrity_check": integrity_check,
                "image_md5s": image_md5_list,
            }

            dat = MultipartEncoder(
                fields={
                    "param": json.dumps(param),
                    "annotated": (aname, open(aname, "rb"), "image/png"),  # image
                    "plom": (pname, open(pname, "rb"), "text/plain"),  # plom-file
                }
            )
            response = self.session.put(
                "https://{}/MK/tasks/{}".format(self.server, code),
                data=dat,
                headers={"Content-Type": dat.content_type},
                verify=False,
            )
            response.raise_for_status()
            ret = response.json()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            elif response.status_code == 406:
                raise PlomConflict(
                    "Integrity check failed. This can happen if manager has altered the task while you are annotating it."
                ) from None
            elif response.status_code == 409:
                raise PlomTaskChangedError("Task ownership has changed.") from None
            elif response.status_code == 410:
                raise PlomTaskDeletedError(
                    "No such task - it has been deleted from server."
                ) from None
            elif response.status_code == 400:
                raise PlomSeriousException(
                    "Image file is corrupted. This should not happen"
                ) from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()
        return ret

    # todo - work out URLs for the various operations a little more nicely.
    def MsetTag(self, code, tags):
        self.SRmutex.acquire()
        try:
            response = self.session.patch(
                "https://{}/MK/tags/{}".format(self.server, code),
                json={"user": self.user, "token": self.token, "tags": tags},
                verify=False,
            )
            response.raise_for_status()

        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            elif response.status_code == 409:
                raise PlomTakenException("Task taken by another user.") from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

    def MrequestWholePaper(self, code, questionNumber=0):
        self.SRmutex.acquire()
        # note - added default value for questionNumber so that this works correctly
        # when called from identifier. - Fixes #921
        try:
            response = self.session.get(
                "https://{}/MK/whole/{}/{}".format(self.server, code, questionNumber),
                json={"user": self.user, "token": self.token},
                verify=False,
            )
            response.raise_for_status()

            # response should be multipart = [ pageData, f1,f2,f3..]
            imagesAsBytes = MultipartDecoder.from_response(response).parts
            images = []
            i = 0
            for iab in imagesAsBytes:
                if i == 0:
                    pageData = json.loads(iab.content)
                else:
                    images.append(
                        BytesIO(iab.content).getvalue()
                    )  # pass back image as bytes
                i += 1

        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            # TODO?
            elif response.status_code == 409:
                raise PlomTakenException("Task taken by another user.") from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()

        return [pageData, images]

    def MrequestWholePaperMetadata(self, code, questionNumber=0):
        """Get metadata about the images in this paper.

        TODO: questionnumber?  why?

        TODO: returns 404, so why not raise that instead?
        """
        self.SRmutex.acquire()
        # note - added default value for questionNumber so that this works correctly
        # when called from identifier. - Fixes #921
        try:
            response = self.session.get(
                "https://{}/MK/TMP/whole/{}/{}".format(
                    self.server, code, questionNumber
                ),
                json={"user": self.user, "token": self.token},
                verify=False,
            )
            response.raise_for_status()
            ret = response.json()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            else:
                raise PlomSeriousException(
                    "Some other sort of error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()
        return ret

    def MrequestOneImage(self, task_code, image_id, md5sum):
        """Download one image from server by its database id.

        args:
            code (str): the task code such as "q1234g9".
                TODO: consider removing code/`task` from URL.
            image_id (int): TODO: int/str?  The key into the server's
                database of images.
            md5sum (str): the expected md5sum, just for sanity checks or
                something I suppose.

        return:
            bytes: png/jpeg or whatever as bytes.

        Errors/Exceptions
            401: not authenticated
            404: no such image
            409: wrong md5sum provided
        """
        self.SRmutex.acquire()
        try:
            response = self.session.get(
                "https://{}/MK/images/{}/{}/{}".format(
                    self.server, task_code, image_id, md5sum
                ),
                json={"user": self.user, "token": self.token},
                verify=False,
            )
            response.raise_for_status()
            image = BytesIO(response.content).getvalue()
        except requests.HTTPError as e:
            if response.status_code == 401:
                raise PlomAuthenticationException() from None
            elif response.status_code == 409:
                raise PlomConflict("Wrong md5sum provided") from None
            elif response.status_code == 404:
                raise PlomNoMoreException("Cannot find image") from None
            else:
                raise PlomSeriousException(
                    "Some other unexpected error {}".format(e)
                ) from None
        finally:
            self.SRmutex.release()
        return image

    # ------------------------
    # ------------------------


from .scanMessenger import ScanMessenger
from .finishMessenger import FinishMessenger
from .managerMessenger import ManagerMessenger



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2018-2020 Andrew Rechnitzer
# Copyright (C) 2020 Colin B. Macdonald

from datetime import datetime
import logging

import peewee as pw

from plom.db.tables import *

log = logging.getLogger("DB")


def createUser(self, uname, passwordHash):
    try:
        uref = User.create(
            name=uname,
            password=passwordHash,
            last_activity=datetime.now(),
            last_action="Created",
        )
    except pw.IntegrityError as e:
        log.error("Create User {} error - {}".format(uname, e))
        return False
    return True


def doesUserExist(self, uname):
    uref = User.get_or_none(name=uname)
    if uref is None:
        return False
    else:
        return True


def setUserPasswordHash(self, uname, passwordHash):
    # Don't mess with HAL
    if uname == "HAL":
        return False
    # token generated by server not DB
    uref = User.get_or_none(name=uname)
    if uref is None:
        return False
    with plomdb.atomic():
        uref.password = passwordHash
        uref.last_activity = datetime.now()
        uref.last_action = "Password set"
        uref.save()
    return True


def getUserPasswordHash(self, uname):
    uref = User.get_or_none(name=uname)
    if uref is None:
        return None
    else:
        return uref.password


def isUserEnabled(self, uname):
    uref = User.get_or_none(name=uname)
    if uref is None:
        return False
    else:
        return uref.enabled


def enableUser(self, uname):
    uref = User.get_or_none(name=uname)
    if uref is None:
        return False
    with plomdb.atomic():
        uref.enabled = True
        uref.save()
    return True


def disableUser(self, uname):
    # when user is disabled we should set the enabled flag to false, remove their auth-token and then remove all their todo-stuff.
    uref = User.get_or_none(name=uname)
    if uref is None:
        return False
    # set enabled flag to false and remove their token
    with plomdb.atomic():
        uref.enabled = False
        uref.token = None
        uref.save()
    # put all of user's tasks back on the todo pile.
    self.resetUsersToDo(uname)
    return True


def setUserToken(self, uname, token, msg="Log on"):
    # token generated by server not DB
    uref = User.get_or_none(name=uname)
    if uref is None:
        return False
    with plomdb.atomic():
        uref.token = token
        uref.last_activity = datetime.now()
        uref.last_action = msg
        uref.save()
    return True


def clearUserToken(self, uname):
    return self.setUserToken(uname, None, "Log off")


def getUserToken(self, uname):
    """Return user's saved token or None if we have no such user."""
    uref = User.get_or_none(name=uname)
    if uref is None:
        return None
    else:
        return uref.token


def userHasToken(self, uname):
    if self.getUserToken(uname) is not None:
        return True
    else:
        return False


def validateToken(self, uname, token):
    uref = User.get_or_none(name=uname)
    if uref is None:
        return False
    if uref.token == token:
        return True
    else:
        return False


def getUserList(self):
    rval = []
    for uref in User.select():
        rval.append(uref.name)
    return rval


def getUserDetails(self):
    rval = {}
    for uref in User.select():
        val = [False, False]
        if uref.enabled:
            val[0] = True
        if uref.token is not None:
            val[1] = True
        if uref.last_activity is None:
            val += ["", ""]
        else:
            val += [
                uref.last_activity.strftime("%y:%m:%d-%H:%M:%S"),
                uref.last_action,
            ]
        rval[uref.name] = val + self.RgetUserFullProgress(uref.name)
    return rval


def resetUsersToDo(self, uname):
    uref = User.get_or_none(name=uname)
    if uref is None:
        return
    with plomdb.atomic():
        query = IDGroup.select().where(IDGroup.user == uref, IDGroup.status == "out")
        for x in query:
            x.status = "todo"
            x.user = None
            x.time = datetime.now()
            x.save()
            log.info("Reset user {} ID task {}".format(uname, x.group.gid))
    with plomdb.atomic():
        query = QGroup.select().where(
            QGroup.user == uref,
            QGroup.status == "out",
        )
        for x in query:
            x.status = "todo"
            x.user = None
            # now clean up the qgroup
            # TODO: why is this code different from db_marks->MdidNotFinish?
            x.save()
            log.info(
                "Reset user {} question-annotation task {}".format(uname, x.group.gid)
            )



__author__ = "Andrew Rechnitzer"
__copyright__ = "Copyright (C) 2020 Andrew Rechnitzer"
__credits__ = ["Andrew Rechnitzer", "Colin Macdonald", "Elvis Cai", "Matt Coles"]
__license__ = "AGPLv3"

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QBrush, QGuiApplication, QPainter, QPixmap
from PyQt5.QtWidgets import (
    QDialog,
    QFrame,
    QFormLayout,
    QGraphicsPixmapItem,
    QGraphicsItemGroup,
    QGraphicsScene,
    QGraphicsView,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSpinBox,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)
from plom.client.examviewwindow import ExamViewWindow


class ActionTab(QWidget):
    def __init__(self, parent):
        super(ActionTab, self).__init__()
        self.parent = parent
        vb = QVBoxLayout()
        self.ob = QPushButton("Keep original page (left)")
        self.cb = QPushButton("Keep colliding page (right)")
        self.vwtb = QPushButton("View whole test")
        vb.addWidget(self.ob)
        vb.addWidget(self.cb)
        vb.addWidget(self.vwtb)
        self.setLayout(vb)
        self.show()
        self.ob.clicked.connect(self.originalPage)
        self.cb.clicked.connect(self.collidingPage)
        self.vwtb.clicked.connect(self.test)

    def originalPage(self):
        self.parent.optionTW.setCurrentIndex(1)
        self.parent.collideGB.setEnabled(False)
        self.parent.collideGB.setStyleSheet("background: red")
        self.parent.originalGB.setStyleSheet("background: green")

    def collidingPage(self):
        self.parent.optionTW.setCurrentIndex(2)
        self.parent.originalGB.setEnabled(False)
        self.parent.collideGB.setStyleSheet("background: green")
        self.parent.originalGB.setStyleSheet("background: red")

    def test(self):
        self.parent.viewWholeTest()


class OriginalTab(QWidget):
    def __init__(self, parent):
        super(OriginalTab, self).__init__()
        self.parent = parent
        vb = QVBoxLayout()
        self.kb = QPushButton("Click to confirm keep original")
        self.ob = QPushButton("Return to other options")
        vb.addStretch(0)
        vb.addWidget(self.kb)
        vb.addStretch(0)
        vb.addWidget(self.ob)
        self.setLayout(vb)
        self.show()
        self.kb.clicked.connect(self.keepOriginal)
        self.ob.clicked.connect(self.other)

    def keepOriginal(self):
        self.parent.action = "original"
        self.parent.accept()

    def other(self):
        self.parent.optionTW.setCurrentIndex(0)
        self.parent.collideGB.setEnabled(True)
        self.parent.collideGB.setStyleSheet("")
        self.parent.originalGB.setStyleSheet("")


class CollideTab(QWidget):
    def __init__(self, parent):
        super(CollideTab, self).__init__()
        self.parent = parent
        vb = QVBoxLayout()
        self.kb = QPushButton("Click to confirm replace")
        self.ob = QPushButton("Return to other options")
        vb.addStretch(0)
        vb.addWidget(self.kb)
        vb.addStretch(0)
        vb.addWidget(self.ob)
        self.setLayout(vb)
        self.show()
        self.kb.clicked.connect(self.keepCollide)
        self.ob.clicked.connect(self.other)

    def keepCollide(self):
        self.parent.action = "collide"
        self.parent.accept()

    def other(self):
        self.parent.optionTW.setCurrentIndex(0)
        self.parent.originalGB.setEnabled(True)
        self.parent.collideGB.setStyleSheet("")
        self.parent.originalGB.setStyleSheet("")


class CollideViewWindow(QDialog):
    """Simple view window for pageimages"""

    def __init__(self, parent, onames, fnames, test, page):
        QWidget.__init__(self)
        self.parent = parent
        self.test = test
        self.page = page

        self.initUI([onames], [fnames])
        self.action = ""

    def initUI(self, onames, fnames):
        self.viewO = ExamViewWindow(onames)
        self.viewC = ExamViewWindow(fnames)

        self.cancelB = QPushButton("&cancel")
        self.maxNormB = QPushButton("&max/norm")

        self.cancelB.clicked.connect(self.reject)
        self.maxNormB.clicked.connect(self.swapMaxNorm)

        self.originalGB = QGroupBox("Original")
        self.collideGB = QGroupBox("Collision")
        self.optionTW = QTabWidget()
        self.initTabs()

        ob = QHBoxLayout()
        ob.addWidget(self.viewO)
        self.originalGB.setLayout(ob)
        cb = QHBoxLayout()
        cb.addWidget(self.viewC)
        self.collideGB.setLayout(cb)

        grid = QGridLayout()
        grid.addWidget(self.maxNormB, 1, 4)
        grid.addWidget(self.cancelB, 10, 4)
        grid.addWidget(self.optionTW, 2, 1, 8, 4)
        self.pane = QWidget()
        self.pane.setLayout(grid)

        hb = QHBoxLayout()
        hb.addWidget(self.originalGB)
        hb.setStretchFactor(self.originalGB, 1)
        hb.addWidget(self.collideGB)
        hb.setStretchFactor(self.collideGB, 1)
        hb.addWidget(self.pane)
        self.setLayout(hb)
        self.show()

    def initTabs(self):
        self.t0 = ActionTab(self)
        self.t1 = OriginalTab(self)
        self.t2 = CollideTab(self)
        self.optionTW.addTab(self.t0, "Actions")
        self.optionTW.addTab(self.t1, "Keep original")
        self.optionTW.addTab(self.t2, "Keep collide")

    def swapMaxNorm(self):
        """Toggles the window size between max and normal"""
        if self.windowState() != Qt.WindowMaximized:
            self.setWindowState(Qt.WindowMaximized)
        else:
            self.setWindowState(Qt.WindowNoState)

    def viewWholeTest(self):
        self.parent.viewWholeTest(self.test)



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2019-2020 Andrew Rechnitzer
# Copyright (C) 2020 Colin B. Macdonald
# Copyright (C) 2020 Vala Vakilian

import csv
import os
from pathlib import Path

from aiohttp import web, MultipartWriter, MultipartReader

from plom import specdir
from .routeutils import authenticate_by_token, authenticate_by_token_required_fields
from .routeutils import log

# I couldn't make this work with the auth deco
# routes = web.RouteTableDef()


class IDHandler:
    def __init__(self, plomServer):
        self.server = plomServer
        # self.local_route_table = routes

    # @routes.get("/ID/progress")
    @authenticate_by_token
    def IDprogressCount(self):
        """Send back current ID progress counts to the client.

        Responds with status 200.

        Returns:
            list: A list of [all the ID'd record , all the records] in the form of int.
        """
        return web.json_response(self.server.IDprogressCount(), status=200)

    # @routes.get("/ID/classlist")
    @authenticate_by_token
    def IDgetClasslist(self):
        """Returns the classlist to the client.

        Used, for example, to fill in the student details for the searchbar autofill.

        Responds with status success or HTTPNotFound.

        The classlist is a ordered list of (student ID, student name)
        pairs.

        Returns:
            class 'aiohttp.web_fileresponse.FileResponse: A file response that includes the classlist.
        """
        try:
            with open(Path(specdir) / "classlist.csv") as csvfile:
                reader = csv.reader(csvfile)
                next(reader)  # skip header row
                classlist = list(reader)
        except FileNotFoundError:
            raise web.HTTPNotFound(reason="classlist not found")
        return web.json_response(classlist)

    # @routes.put("/ID/classlist")
    @authenticate_by_token_required_fields(["user", "classlist"])
    def IDputClasslist(self, data, request):
        """Accept classlist upload.

        Only "manager" can perform this action.
        Responds with status success, HTTPBadRequest or HTTPConflict.

        The classlist should be provided as a ordered list of (str, str)
        pairs where each pair is (student ID, student name).

        Returns:
            aiohttp.web_response.Response: A response indicating whether the operation was a failure or a success.
        """
        if not data["user"] == "manager":
            raise web.HTTPBadRequest(reason="Not manager")
        classlist = data["classlist"]
        if os.path.isfile(Path(specdir) / "classlist.csv"):
            raise web.HTTPConflict(reason="we already have a classlist")
        with open(Path(specdir) / "classlist.csv", "w") as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow(["id", "studentName"])
            writer.writerows(classlist)
        return web.Response()

    # @routes.get("/ID/predictions")
    @authenticate_by_token
    def IDgetPredictions(self):
        """Returns the files involving the ML model's student id's prediction.

        Responds with status 200/404.

        Returns:
            aiohttp.web_fileresponse.FileResponse: File response including the predictions.
        """
        if os.path.isfile(Path(specdir) / "predictionlist.csv"):
            return web.FileResponse(Path(specdir) / "predictionlist.csv", status=200)
        else:
            return web.Response(status=404)

    # @routes.get("/ID/tasks/complete")
    @authenticate_by_token_required_fields(["user"])
    def IDgetDoneTasks(self, data, request):
        """Responds with a list of id/name which have already been confirmed by the client.

        Args:
            data (dict): A (str:str) dictionary having keys `user` and `token`.
            request (aiohttp.web_request.Request): GET /ID/tasks/complete  request type.

        Returns:
            aiohttp.web_request.Request: A response including a list of lists indicating information about
                the users who already have confirmed predictions.
                Each list in the response is of the format: [task_number, task_status, student_id, student_name]
        """

        # return the completed list
        return web.json_response(self.server.IDgetDoneTasks(data["user"]), status=200)

    # @routes.get("/ID/images/{test}")
    @authenticate_by_token_required_fields(["user"])
    def IDgetImage(self, data, request):
        """Return the images for a specified paper number.

        Responds with status 200/404/409/410.

        Args:
            data (dict): A (str:str) dictionary having keys `user` and `token`.
            request (aiohttp.web_request.Request): DELETE /ID/predictedID type request object.

        Returns:
            aiohttp.web_fileresponse.FileResponse: A response including a aiohttp object which
                includes a multipart object with the images.
        """

        # A string having the sting number.
        test_number = request.match_info["test"]

        image_path = self.server.IDgetImage(data["user"], test_number)
        # is either user allowed access - returns [true, fname0, fname1,...]
        # or fails - return [false, message]

        allow_access = image_path[0]
        if not allow_access:
            # can fail for 3 reasons - "not owner", "no such scan", "no such test"
            fail_message = image_path[1]
            if fail_message == "NotOwner":
                return web.Response(status=409)  # someone else has that image
            elif fail_message == "NoScan":
                return web.Response(status=410)
            else:  # fail_message == "NoTest":
                return web.Response(status=404)

        with MultipartWriter("images") as writer:
            image_paths = image_path[1:]

            for file_name in image_paths:
                if os.path.isfile(file_name):
                    writer.append(open(file_name, "rb"))
                else:
                    return web.Response(status=404)
            return web.Response(body=writer, status=200)

    # @routes.get("/ID/tasks/available")
    @authenticate_by_token
    def IDgetNextTask(self):
        """Responds with a code for the the next available identify task.

        Note: There is no guarantee that task will still be available later but at this moment in time,
        no one else has claimed it

        Responds with status 200/204.

        Returns:
            aiohttp.web_response.Response: A response object with the code for the next task/paper.
        """

        next_task_code = self.server.IDgetNextTask()  # returns [True, code] or [False]
        next_task_available = next_task_code[0]

        if next_task_available:
            next_task_code = next_task_code[1]
            return web.json_response(next_task_code, status=200)
        else:
            return web.Response(status=204)  # no papers left

    # @routes.patch("/ID/tasks/{task}")
    @authenticate_by_token_required_fields(["user"])
    def IDclaimThisTask(self, data, request):
        """Claims this identifying task and returns images of the ID pages.

        Responds with status 200/204/404.

        Args:
            data (dict): A (str:str) dictionary having keys `user` and `token`.
            request (aiohttp.web_request.Request): PATCH /ID/tasks request object.

        Returns:
            aiohttp.web_response.Response: A response including a aiohttp object which
                includes a multipart object with the images.
        """

        testNumber = request.match_info["task"]
        image_path = self.server.IDclaimThisTask(
            data["user"], testNumber
        )  # returns [True, IMG_path] or [False]

        allow_access = image_path[0]

        if allow_access:  # user allowed access - returns [true, fname0, fname1,...]
            with MultipartWriter("images") as writer:
                image_paths = image_path[1:]

                for file_name in image_paths:
                    if os.path.isfile(file_name):
                        writer.append(open(file_name, "rb"))
                    else:
                        return web.Response(status=404)
                return web.Response(body=writer, status=200)
        else:
            return web.Response(status=204)  # that task already taken.

    # @routes.put("/ID/tasks/{task}")
    @authenticate_by_token_required_fields(["user", "sid", "sname"])
    def IdentifyPaperTask(self, data, request):
        """Identify a paper based on a task.

        Returns:
            403: some other user owns this task.
            404: papernum not found, or other data errors.
            409: student number `data["sid"]` is already in use.
        """
        papernum = request.match_info["task"]
        r, what, msg = self.server.ID_id_paper(
            papernum, data["user"], data["sid"], data["sname"]
        )
        if r:
            return web.Response(status=200)
        elif what == 409:
            raise web.HTTPConflict(reason=msg)
        elif what == 404:
            raise web.HTTPNotFound(reason=msg)
        elif what == 403:
            raise web.HTTPForbidden(reason=msg)
        else:
            # catch all that should not happen.
            raise web.HTTPInternalServerError(reason=msg)

    # @routes.put("/ID/{papernum}")
    @authenticate_by_token_required_fields(["user", "sid", "sname"])
    def IdentifyPaper(self, data, request):
        """Identify a paper directly without certain checks.

        Only "manager" can perform this action.  Typical client IDing
        would call func:`IdentifyPaperTask` instead.

        Returns:
            400: not manager.
            404: papernum not found, or other data errors.
            409: student number `data["sid"]` is already in use.
        """
        if not data["user"] == "manager":
            raise web.HTTPBadRequest(reason="Not manager")
        papernum = request.match_info["papernum"]
        r, what, msg = self.server.id_paper(papernum, "HAL", data["sid"], data["sname"])
        if r:
            return web.Response(status=200)
        elif what == 409:
            raise web.HTTPConflict(reason=msg)
        elif what == 404:
            raise web.HTTPNotFound(reason=msg)
        else:
            # catch all that should not happen.
            raise web.HTTPInternalServerError(reason=msg)

    # @routes.delete("/ID/tasks/{task}")
    @authenticate_by_token_required_fields(["user"])
    def IDdidNotFinishTask(self, data, request):
        """Accept the client's surrender of a previously-claimed identifying task.

        This could occur for example when the client closes with unfinished tasks.
        Responds with status 200.

        Args:
            data (dict): A (str:str) dictionary having keys `user` and `token`.
            request (aiohttp.web_request.Request'): Request of type DELETE /ID/tasks/#TaskNumber.

        Returns:
            aiohttp.web_response.Response: A response with status 200.
        """

        testNumber = request.match_info["task"]
        self.server.IDdidNotFinish(data["user"], testNumber)
        return web.json_response(status=200)

    # @routes.get("/ID/randomImage")
    @authenticate_by_token_required_fields(["user"])
    def IDgetImageFromATest(self, data, request):
        """Gets a random image to extract the bounding box corresponding to the student name and id.

        The bounding box indicated on this image will be later used to extract the
        student ids from the other papers.
        Responds with status 200/404/410.
        Logs activity.

        Args:
            data (dict): A (str:str) dictionary having keys `user` and `token`.
            request (aiohttp.web_request.Request): request of type GET /ID/randomImage.
        Returns:
            aiohttp.web_fileresponse.FileResponse: A response including a aiohttp object which
                includes a multipart object with the images.
        """

        # TODO: maybe we want some special message here?
        if data["user"] != "manager":
            return web.Response(status=401)  # only manager

        # A list with a boolean (indicating wether the objects exist) and a list of the exam images.
        random_image_paths = self.server.IDgetImageFromATest()

        allow_access = random_image_paths[0]

        # No access to the files
        if allow_access is False:
            return web.Response(status=410)

        log.debug("Appending file {}".format(random_image_paths))
        with MultipartWriter("images") as writer:
            image_paths = random_image_paths[1:]

            for file_name in image_paths:
                if os.path.isfile(file_name):
                    writer.append(open(file_name, "rb"))
                else:
                    return web.Response(status=404)
            return web.Response(body=writer, status=200)

    @authenticate_by_token_required_fields(["user"])
    def IDdeletePredictions(self, data, request):
        """Deletes the machine-learning predicted IDs for all papers.

        Responds with status 200/401.

        Args:
            data (dict): A (str:str) dictionary having keys `user` and `token`.
            request (aiohttp.web_request.Request): DELETE /ID/predictedID type request object.

        Returns:
            aiohttp.web_response.Response: Returns a response with a True or False indicating if the deletion
                was successful.
        """

        # TODO: maybe we want some special message here?
        if data["user"] != "manager":
            return web.Response(status=401)

        return web.json_response(self.server.IDdeletePredictions(), status=200)

    @authenticate_by_token_required_fields(
        ["user", "rectangle", "fileNumber", "ignoreStamp"]
    )
    def IDrunPredictions(self, data, request):
        """Runs the id prediction on all paper images.

        Responds with status 200/202/205/401.

        Args:
            data (dict): A dictionary having the user/token in addition to information from the rectangle
                bounding box coordinates and file information.
            request (aiohttp.web_request.Request): Request of type POST /ID/predictedID.

        Returns:
            aiohttp.web_response.Response: Returns a response with the date and time of the prediction run.
                Or responds with saying the prediction is already running.
        """

        # TODO: maybe we want some special message here?
        if data["user"] != "manager":
            return web.Response(status=401)

        prediction_results = self.server.IDrunPredictions(
            data["rectangle"], data["fileNumber"], data["ignoreStamp"]
        )

        timestamp_found = prediction_results[0]
        is_running = prediction_results[1]

        if timestamp_found:  # set running or is running
            if is_running:
                return web.Response(status=200)
            else:
                return web.Response(status=202)  # is already running
        else:  # isn't running because we found a time-stamp
            return web.Response(text=is_running, status=205)

    # @routes.patch("/ID/review")
    @authenticate_by_token_required_fields(["testNumber"])
    def IDreviewID(self, data, request):
        """Responds with an empty response object indicating if the review ID is possible and the document exists.

        Responds with status 200/404.

        Args:
            data (dict): A dictionary having the user/token in addition to the `testNumber`.
            request (aiohttp.web_request.Request): Request of type PATCH /ID/review.

        Returns:
            aiohttp.web_fileresponse.FileResponse: An empty response indicating the availability status of
                the review document.
        """

        if self.server.IDreviewID(data["testNumber"]):
            return web.Response(status=200)
        else:
            return web.Response(status=404)

    def setUpRoutes(self, router):
        """Adds the response functions to the router object.

        Args:
            router (aiohttp.web_urldispatcher.UrlDispatcher): Router object which we will add the response functions to.
        """

        # router.add_routes(self.local_route_table)
        # But see above: doesn't work with auth deco
        router.add_get("/ID/progress", self.IDprogressCount)
        router.add_get("/ID/classlist", self.IDgetClasslist)
        router.add_put("/ID/classlist", self.IDputClasslist)
        router.add_get("/ID/predictions", self.IDgetPredictions)
        router.add_get("/ID/tasks/complete", self.IDgetDoneTasks)
        router.add_get("/ID/images/{test}", self.IDgetImage)
        router.add_get("/ID/tasks/available", self.IDgetNextTask)
        router.add_patch("/ID/tasks/{task}", self.IDclaimThisTask)
        router.add_put("/ID/{papernum}", self.IdentifyPaper)
        router.add_put("/ID/tasks/{task}", self.IdentifyPaperTask)
        router.add_delete("/ID/tasks/{task}", self.IDdidNotFinishTask)
        router.add_get("/ID/randomImage", self.IDgetImageFromATest)
        router.add_delete("/ID/predictedID", self.IDdeletePredictions)
        router.add_post("/ID/predictedID", self.IDrunPredictions)
        router.add_patch("/ID/review", self.IDreviewID)



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2018-2020 Andrew Rechnitzer
# Copyright (C) 2020 Colin B. Macdonald
# Copyright (C) 2020 Victoria Schuster

from PyQt5.QtCore import Qt, QPointF, QTimer, QPropertyAnimation, pyqtProperty
from PyQt5.QtGui import QFont, QImage, QPen, QColor, QBrush
from PyQt5.QtWidgets import QUndoCommand, QGraphicsItem, QGraphicsTextItem


class CommandMoveText(QUndoCommand):
    # Moves the textitem. we give it an ID so it can be merged with other
    # commandmoves on the undo-stack.
    # Don't use this for moving other graphics items
    # Graphicsitems are separate from graphicsTEXTitems
    def __init__(self, xitem, new_pos):
        super(CommandMoveText, self).__init__()
        self.xitem = xitem
        self.old_pos = xitem.pos()
        self.new_pos = new_pos
        self.setText("MoveText")

    def id(self):
        # Give it an id number for merging of undo/redo commands
        return 102

    def redo(self):
        # Temporarily disable the item emiting "I've changed" signals
        self.xitem.setFlag(QGraphicsItem.ItemSendsGeometryChanges, False)
        # Move the object
        self.xitem.setPos(self.new_pos)
        # Reenable the item emiting "I've changed" signals
        self.xitem.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)

    def undo(self):
        # Temporarily disable the item emiting "I've changed" signals
        self.xitem.setFlag(QGraphicsItem.ItemSendsGeometryChanges, False)
        # Move the object back
        self.xitem.setPos(self.old_pos)
        # Reenable the item emiting "I've changed" signals
        self.xitem.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)

    def mergeWith(self, other):
        # Most commands cannot be merged - make sure the moved items are the
        # same - if so then merge things.
        if self.xitem != other.xitem:
            return False
        self.new_pos = other.new_pos
        return True


class CommandText(QUndoCommand):
    def __init__(self, scene, blurb, ink):
        super(CommandText, self).__init__()
        self.scene = scene
        # set no interaction on scene's textitem - this avoids button-mashing
        # issues where text can be added during pasting in of text
        iflags = blurb.textInteractionFlags()
        blurb.setTextInteractionFlags(Qt.NoTextInteraction)
        # copy that textitem to one for this comment
        self.blurb = blurb
        # set the interaction flags back
        blurb.setTextInteractionFlags(iflags)
        # if the textitem has contents already then we
        # have to do some cleanup - give it focus and then
        # drop focus - correctly sets the text interaction flags
        if len(self.blurb.toPlainText()) > 0:
            QTimer.singleShot(1, self.blurb.setFocus)
            QTimer.singleShot(2, self.blurb.clearFocus)
            QTimer.singleShot(5, self.blurb.textToPng)
        self.setText("Text")

    @classmethod
    def from_pickle(cls, X, *, scene):
        """Construct a CommandText from a serialized form."""
        assert X[0] == "Text"
        X = X[1:]
        if len(X) != 3:
            raise ValueError("wrong length of pickle data")
        blurb = TextItem(scene, scene.fontSize)
        blurb.setPlainText(X[0])
        blurb._contents = X[0]  # TODO
        blurb.setPos(QPointF(X[1], X[2]))
        blurb.setTextInteractionFlags(Qt.NoTextInteraction)
        # knows to latex it if needed.
        return cls(scene, blurb, scene.ink)

    def redo(self):
        self.blurb.flash_redo()
        self.scene.addItem(self.blurb)

    def undo(self):
        self.blurb.flash_undo()
        QTimer.singleShot(200, lambda: self.scene.removeItem(self.blurb))


class TextItem(QGraphicsTextItem):
    # Textitem is a qgraphicstextitem, has to handle
    # textinput and double-click to start editing etc.
    # Shift-return ends the editor
    def __init__(self, parent, fontsize=10):
        super(TextItem, self).__init__()
        self.saveable = True
        self.animator = [self]
        self.animateFlag = False
        self.parent = parent
        # Thick is thickness of bounding box hightlight used
        # to highlight the object when undo / redo happens.
        self.thick = 0
        self.setDefaultTextColor(Qt.red)
        self.setPlainText("")
        self._contents = ""
        font = QFont("Helvetica")
        font.setPointSizeF(fontsize)
        self.setFont(font)
        self.setFlag(QGraphicsItem.ItemIsMovable)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges)
        # Set it as editably with the text-editor
        self.setTextInteractionFlags(Qt.TextEditorInteraction)
        # Undo/redo animates via the thickness property
        self.anim = QPropertyAnimation(self, b"thickness")
        # for latex png
        self.state = "TXT"

    # TODO: override toPlainText() to behave more like the super class
    # def toPlainText():

    def getContents(self):
        # TODO: several different ways to check this: consolidate
        if len(self._contents) == 0:
            return self.toPlainText()
        else:
            return self._contents

    def focusInEvent(self, event):
        if self.state == "PNG":
            self.pngToText()
        else:
            self._contents = self.toPlainText()
        super(TextItem, self).focusInEvent(event)

    def focusOutEvent(self, event):
        # When object loses the focus, need to make sure that
        # the editor stops, any highlighted text is released
        # and stops taking any text-interactions.
        tc = self.textCursor()
        tc.clearSelection()
        self.setTextCursor(tc)
        self.setTextInteractionFlags(Qt.NoTextInteraction)
        # if not PNG then update contents
        if self.state != "PNG":
            self._contents = self.toPlainText()
        super(TextItem, self).focusOutEvent(event)

    def textToPng(self):
        self._contents = self.toPlainText()
        if self._contents[:4].upper() == "TEX:":
            texIt = self._contents[4:]
        else:
            # is not latex so we don't have to PNG-it
            return

        fragfilename = self.parent.latexAFragment(texIt)
        if fragfilename:
            self.setPlainText("")
            tc = self.textCursor()
            qi = QImage(fragfilename)
            tc.insertImage(qi)
            self.state = "PNG"

    def pngToText(self):
        # TODO: several different ways to check this: consolidate
        if self._contents != "":
            self.setPlainText(self._contents)
        self.state = "TXT"

    def keyPressEvent(self, event):
        # Shift-return ends the editor and releases the object
        if event.modifiers() == Qt.ShiftModifier and event.key() == Qt.Key_Return:
            # Clear any highlighted text and release.
            tc = self.textCursor()
            tc.clearSelection()
            self.setTextCursor(tc)
            self.setTextInteractionFlags(Qt.NoTextInteraction)
            self._contents = self.toPlainText()
            if self._contents[:4].upper() == "TEX:":
                self.textToPng()

        # TODO: I don't understand how this differs from shift-enter?
        # TODO: was ctrl-enter supposed to latex even without the "tex:"?
        # control-return latexs the comment and replaces the text with the resulting image.
        # ends the editor.
        if event.modifiers() == Qt.ControlModifier and event.key() == Qt.Key_Return:
            self.textToPng()
            tc = self.textCursor()
            tc.clearSelection()
            self.setTextCursor(tc)
            self.setTextInteractionFlags(Qt.NoTextInteraction)

        super(TextItem, self).keyPressEvent(event)

    def paint(self, painter, option, widget):
        if not self.scene().itemWithinBounds(self):
            if self.group() is None:  # make sure not part of a GDT
                painter.setPen(QPen(QColor(255, 165, 0), 8))
                painter.setBrush(QBrush(QColor(255, 165, 0, 128)))
                painter.drawLine(option.rect.topLeft(), option.rect.bottomRight())
                painter.drawLine(option.rect.topRight(), option.rect.bottomLeft())
                painter.drawRoundedRect(option.rect, 10, 10)
        else:
            # paint a bounding rectangle for undo/redo highlighting
            if self.thick > 0:
                painter.setPen(QPen(Qt.red, self.thick))
                painter.drawRoundedRect(option.rect, 10, 10)
        # paint the normal TextItem with the default 'paint' method
        super(TextItem, self).paint(painter, option, widget)

    def itemChange(self, change, value):
        if change == QGraphicsTextItem.ItemPositionChange and self.scene():
            command = CommandMoveText(self, value)
            # Notice that the value here is the new position, not the delta.
            self.scene().undoStack.push(command)
        return QGraphicsTextItem.itemChange(self, change, value)

    def flash_undo(self):
        # When undo-ing, draw a none->thick->none border around text.
        self.anim.setDuration(200)
        self.anim.setStartValue(0)
        self.anim.setKeyValueAt(0.5, 8)
        self.anim.setEndValue(0)
        self.anim.start()

    def flash_redo(self):
        # When redo-ing, draw a none->med->none border around text.
        self.anim.setDuration(200)
        self.anim.setStartValue(0)
        self.anim.setKeyValueAt(0.5, 4)
        self.anim.setEndValue(0)
        self.anim.start()

    def pickle(self):
        # TODO: several different ways to check this: consolidate
        if len(self._contents) == 0:
            self._contents = self.toPlainText()
        return ["Text", self._contents, self.scenePos().x(), self.scenePos().y()]

    # For the animation of border
    @pyqtProperty(int)
    def thickness(self):
        return self.thick

    # For the animation of border
    @thickness.setter
    def thickness(self, value):
        self.thick = value
        self.update()


class GhostText(QGraphicsTextItem):
    # Textitem is a qgraphicstextitem, has to handle
    # textinput and double-click to start editing etc.
    # Shift-return ends the editor
    def __init__(self, txt, fontsize=10):
        super(GhostText, self).__init__()
        self.setDefaultTextColor(Qt.blue)
        self.setPlainText(txt)
        font = QFont("Helvetica")
        font.setPointSizeF(fontsize)
        self.setFont(font)
        self.setFlag(QGraphicsItem.ItemIsMovable)
        # Set it as editably with the text-editor
        self.setTextInteractionFlags(Qt.NoTextInteraction)
        # If we're displaying png-rendered-latex then we will store the
        # original text here
        self._png_tex_cache = None

    def is_displaying_png(self):
        """Is this TextItem displaying a PNG, e.g., of LaTeX?"""
        return not self._png_tex_cache is None

    def changeText(self, txt):
        self._png_tex_cache = None
        self.setPlainText(txt)
        if self.scene() is not None and txt[:4].upper() == "TEX:":
            texIt = (
                "\\color{blue}\n" + txt[4:].strip()
            )  # make color blue for ghost rendering
            fragfilename = self.scene().latexAFragment(texIt)
            if fragfilename:
                self._png_tex_cache = txt
                self.setPlainText("")
                tc = self.textCursor()
                qi = QImage(fragfilename)
                tc.insertImage(qi)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'qtCreatorFiles/ui_marker.ui'
#
# Created by: PyQt5 UI code generator 5.14.2
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MarkerWindow(object):
    def setupUi(self, MarkerWindow):
        MarkerWindow.setObjectName("MarkerWindow")
        MarkerWindow.setWindowModality(QtCore.Qt.WindowModal)
        MarkerWindow.resize(1024, 768)
        self.horizontalLayout = QtWidgets.QHBoxLayout(MarkerWindow)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.splitter = QtWidgets.QSplitter(MarkerWindow)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setHandleWidth(10)
        self.splitter.setObjectName("splitter")
        self.frame = QtWidgets.QFrame(self.splitter)
        self.frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame.setObjectName("frame")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.frame)
        self.verticalLayout.setContentsMargins(0, 0, 0, -1)
        self.verticalLayout.setObjectName("verticalLayout")
        self.infoBox = QtWidgets.QGroupBox(self.frame)
        self.infoBox.setObjectName("infoBox")
        self.formLayout_2 = QtWidgets.QFormLayout(self.infoBox)
        self.formLayout_2.setObjectName("formLayout_2")
        self.label_2 = QtWidgets.QLabel(self.infoBox)
        self.label_2.setObjectName("label_2")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.userLabel = QtWidgets.QLabel(self.infoBox)
        self.userLabel.setText("<username>")
        self.userLabel.setObjectName("userLabel")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.userLabel)
        self.label_5 = QtWidgets.QLabel(self.infoBox)
        self.label_5.setObjectName("label_5")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_5)
        self.maxscoreLabel = QtWidgets.QLabel(self.infoBox)
        self.maxscoreLabel.setText("<number>")
        self.maxscoreLabel.setObjectName("maxscoreLabel")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.maxscoreLabel)
        self.verticalLayout.addWidget(self.infoBox)
        self.tableBox = QtWidgets.QGroupBox(self.frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.tableBox.sizePolicy().hasHeightForWidth())
        self.tableBox.setSizePolicy(sizePolicy)
        self.tableBox.setObjectName("tableBox")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.tableBox)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.tableView = SimpleTableView(self.tableBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tableView.sizePolicy().hasHeightForWidth())
        self.tableView.setSizePolicy(sizePolicy)
        self.tableView.setMinimumSize(QtCore.QSize(250, 0))
        self.tableView.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.tableView.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.tableView.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.tableView.setObjectName("tableView")
        self.verticalLayout_2.addWidget(self.tableView)
        self.frame1 = QtWidgets.QFrame(self.tableBox)
        self.frame1.setObjectName("frame1")
        self.gridLayout = QtWidgets.QGridLayout(self.frame1)
        self.gridLayout.setContentsMargins(0, 1, 0, 0)
        self.gridLayout.setObjectName("gridLayout")
        self.annButton = QtWidgets.QPushButton(self.frame1)
        self.annButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.annButton.setObjectName("annButton")
        self.gridLayout.addWidget(self.annButton, 0, 0, 1, 2)
        self.getNextButton = QtWidgets.QPushButton(self.frame1)
        self.getNextButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.getNextButton.setObjectName("getNextButton")
        self.gridLayout.addWidget(self.getNextButton, 0, 2, 1, 1)
        self.deferButton = QtWidgets.QPushButton(self.frame1)
        self.deferButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.deferButton.setObjectName("deferButton")
        self.gridLayout.addWidget(self.deferButton, 1, 0, 1, 1)
        self.tagButton = QtWidgets.QPushButton(self.frame1)
        self.tagButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.tagButton.setObjectName("tagButton")
        self.gridLayout.addWidget(self.tagButton, 1, 1, 1, 1)
        self.viewButton = QtWidgets.QPushButton(self.frame1)
        self.viewButton.setObjectName("viewButton")
        self.gridLayout.addWidget(self.viewButton, 1, 2, 1, 1)
        self.verticalLayout_2.addWidget(self.frame1)
        self.frame2 = QtWidgets.QFrame(self.tableBox)
        self.frame2.setObjectName("frame2")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.frame2)
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.filterButton = QtWidgets.QPushButton(self.frame2)
        self.filterButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.filterButton.setObjectName("filterButton")
        self.horizontalLayout_2.addWidget(self.filterButton)
        self.filterInvCB = QtWidgets.QCheckBox(self.frame2)
        self.filterInvCB.setObjectName("filterInvCB")
        self.horizontalLayout_2.addWidget(self.filterInvCB)
        self.filterLE = QtWidgets.QLineEdit(self.frame2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.filterLE.sizePolicy().hasHeightForWidth())
        self.filterLE.setSizePolicy(sizePolicy)
        self.filterLE.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.filterLE.setMaxLength(256)
        self.filterLE.setClearButtonEnabled(True)
        self.filterLE.setObjectName("filterLE")
        self.horizontalLayout_2.addWidget(self.filterLE)
        self.verticalLayout_2.addWidget(self.frame2)
        self.verticalLayout.addWidget(self.tableBox)
        self.OptionsBox = QtWidgets.QGroupBox(self.frame)
        self.OptionsBox.setObjectName("OptionsBox")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.OptionsBox)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.markUpRB = QtWidgets.QRadioButton(self.OptionsBox)
        self.markUpRB.setChecked(True)
        self.markUpRB.setObjectName("markUpRB")
        self.markStyleGroup = QtWidgets.QButtonGroup(MarkerWindow)
        self.markStyleGroup.setObjectName("markStyleGroup")
        self.markStyleGroup.addButton(self.markUpRB)
        self.verticalLayout_3.addWidget(self.markUpRB)
        self.markDownRB = QtWidgets.QRadioButton(self.OptionsBox)
        self.markDownRB.setChecked(False)
        self.markDownRB.setObjectName("markDownRB")
        self.markStyleGroup.addButton(self.markDownRB)
        self.verticalLayout_3.addWidget(self.markDownRB)
        self.markTotalRB = QtWidgets.QRadioButton(self.OptionsBox)
        self.markTotalRB.setObjectName("markTotalRB")
        self.markStyleGroup.addButton(self.markTotalRB)
        self.verticalLayout_3.addWidget(self.markTotalRB)
        self.leftMouseCB = QtWidgets.QCheckBox(self.OptionsBox)
        self.leftMouseCB.setObjectName("leftMouseCB")
        self.verticalLayout_3.addWidget(self.leftMouseCB)
        self.sidebarRightCB = QtWidgets.QCheckBox(self.OptionsBox)
        self.sidebarRightCB.setObjectName("sidebarRightCB")
        self.verticalLayout_3.addWidget(self.sidebarRightCB)
        self.verticalLayout.addWidget(self.OptionsBox)
        self.frameProgress = QtWidgets.QFrame(self.frame)
        self.frameProgress.setObjectName("frameProgress")
        self.layoutProgress = QtWidgets.QHBoxLayout(self.frameProgress)
        self.layoutProgress.setContentsMargins(0, -1, 0, -1)
        self.layoutProgress.setObjectName("layoutProgress")
        self.labelProgress = QtWidgets.QLabel(self.frameProgress)
        self.labelProgress.setObjectName("labelProgress")
        self.layoutProgress.addWidget(self.labelProgress)
        self.mProgressBar = QtWidgets.QProgressBar(self.frameProgress)
        self.mProgressBar.setProperty("value", 1)
        self.mProgressBar.setObjectName("mProgressBar")
        self.layoutProgress.addWidget(self.mProgressBar)
        self.verticalLayout.addWidget(self.frameProgress)
        self.frameClose = QtWidgets.QFrame(self.frame)
        self.frameClose.setObjectName("frameClose")
        self.layoutClose = QtWidgets.QHBoxLayout(self.frameClose)
        self.layoutClose.setContentsMargins(0, 0, 0, 0)
        self.layoutClose.setObjectName("layoutClose")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.layoutClose.addItem(spacerItem)
        self.closeButton = QtWidgets.QPushButton(self.frameClose)
        self.closeButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.closeButton.setObjectName("closeButton")
        self.layoutClose.addWidget(self.closeButton)
        self.verticalLayout.addWidget(self.frameClose)
        self.paperBox = QtWidgets.QGroupBox(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.paperBox.sizePolicy().hasHeightForWidth())
        self.paperBox.setSizePolicy(sizePolicy)
        self.paperBox.setObjectName("paperBox")
        self.gridLayout_6 = QtWidgets.QGridLayout(self.paperBox)
        self.gridLayout_6.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_6.setObjectName("gridLayout_6")
        self.horizontalLayout.addWidget(self.splitter)

        self.retranslateUi(MarkerWindow)
        QtCore.QMetaObject.connectSlotsByName(MarkerWindow)
        MarkerWindow.setTabOrder(self.tableView, self.closeButton)

    def retranslateUi(self, MarkerWindow):
        _translate = QtCore.QCoreApplication.translate
        MarkerWindow.setWindowTitle(_translate("MarkerWindow", "Mark papers"))
        self.infoBox.setTitle(_translate("MarkerWindow", "Question Info"))
        self.label_2.setText(_translate("MarkerWindow", "Username "))
        self.label_5.setText(_translate("MarkerWindow", "Max score"))
        self.tableBox.setTitle(_translate("MarkerWindow", "Paper list"))
        self.annButton.setText(_translate("MarkerWindow", "&Annotate && mark"))
        self.getNextButton.setText(_translate("MarkerWindow", "&Get next"))
        self.deferButton.setText(_translate("MarkerWindow", "Defer"))
        self.tagButton.setText(_translate("MarkerWindow", "Tag"))
        self.viewButton.setText(_translate("MarkerWindow", "View"))
        self.filterButton.setText(_translate("MarkerWindow", "Filter"))
        self.filterInvCB.setText(_translate("MarkerWindow", "Inv"))
        self.filterLE.setPlaceholderText(_translate("MarkerWindow", "Filter on tag text"))
        self.OptionsBox.setTitle(_translate("MarkerWindow", "Options"))
        self.markUpRB.setText(_translate("MarkerWindow", "Mark &Up style"))
        self.markDownRB.setText(_translate("MarkerWindow", "Mark &Down style"))
        self.markTotalRB.setText(_translate("MarkerWindow", "&Total style"))
        self.leftMouseCB.setToolTip(_translate("MarkerWindow", "Move keyboard shortcuts to right hand"))
        self.leftMouseCB.setText(_translate("MarkerWindow", "&Left-handed mouse"))
        self.sidebarRightCB.setText(_translate("MarkerWindow", "&Sidebar on right"))
        self.labelProgress.setText(_translate("MarkerWindow", "Progress:"))
        self.mProgressBar.setFormat(_translate("MarkerWindow", "%v of %m"))
        self.closeButton.setText(_translate("MarkerWindow", "&Close"))
        self.paperBox.setTitle(_translate("MarkerWindow", "Current paper"))
from plom.client.useful_classes import SimpleTableView



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2019-2020 Colin B. Macdonald
# Copyright (C) 2020 Andrew Rechnitzer

"""Utilities for dealing with TPV codes

A TPV code is a string of 17 digits of the form

  EETTTTPPVVOCCCCCC
  01234567890123456

This is as much as can be packed into a QR code without
increasing its size.

The format consists of:
  * 01 = API
  * 2345 = test number
  * 67 = page number
  * 89 = version
  * 0 = position/orientation code
  * 123456 = magic code

A TPV code can be optionally prefixed with "QR-Code:"

  QR-Code:EETTTTPPVVOCCCCCC

Orientation code:
  * 0 no position information (reserved, unused)
  * 1 NE
  * 2 NW
  * 3 SW
  * 4 SE
  * 5-9 (currently unused)

TODO: might encapsulate position code, have ``getPosition`` return
e.g., the string `"NE"`
"""

import random


# Changes to this format should bump this.  Possibly changes
# to the layout of QR codes on the page should too.
_API = "00"


def isValidTPV(tpv):
    """Is this a valid TPV code?"""
    tpv = tpv.lstrip("QR-Code:")
    if len(tpv) != len("EETTTTPPVVOCCCCCC"):
        return False
    return tpv.isnumeric()


def parseTPV(tpv):
    """Parse a TPV string (typically from a QR-code)

    Args: tpv (str): a TPV string of the form "EETTTTPPVVOCCCCCC",
       typically from a QR-code, possibly with the prefix "QR-Code:".

    Returns:
       tn (int): test number, up to 4 digits
       pn (int): page group number, up to 2 digits
       vn (int): version number, up to 2 digits
       en (str): the API number, 2 digits zero padded
       cn (str): the "magic code", 6 digits zero padded
       o (str): the orientation code, TODO
    """
    tpv = tpv.lstrip("QR-Code:")
    # en = tpv[0:2]
    tn = int(tpv[2:6])
    pn = int(tpv[6:8])
    vn = int(tpv[8:10])
    o = tpv[10]
    cn = tpv[11:]
    return tn, pn, vn, cn, o


def getPosition(tpv):
    return int(parseTPV(tpv)[4])


def getAPI(tpv):
    tpv = tpv.lstrip("QR-Code:")
    return tpv[0:2]


def hasCurrentAPI(tpv):
    """does tpv have the current API?

    Args: tpv (str): a TPV string of the form "EETTTTPPVVOCCCCCC",
       typically from a QR-code.
    """
    return getAPI(tpv) == _API


def getCode(tpv):
    """return the magic code for tpv

    Args: tpv (str): a TPV string.
    """
    return parseTPV(tpv)[3]


def encodeTPV(test, p, v, o, code):
    """Encode some values as a TPV code

    Args:
       test (int/str): the test number, 1 ≤ test ≤ 9999
       p (int/str): page number, 1 ≤ p ≤ 99
       v (int/str): version number, 1 ≤ v ≤ 99
       o (int/str): position code, 0 ≤ code ≤ 4
       code (int/str): magic code, 0 ≤ code ≤ 999999

    Returns:
       str: the tpv code
    """
    assert int(test) >= 1
    assert int(v) >= 1
    assert int(p) >= 1
    assert int(o) >= 0 and int(o) <= 4
    test = str(test).zfill(4)
    p = str(p).zfill(2)
    v = str(v).zfill(2)
    o = str(o)
    code = str(code).zfill(6)
    assert test.isnumeric()
    assert p.isnumeric()
    assert v.isnumeric()
    assert o.isnumeric()
    assert code.isnumeric()
    assert len(test) == 4
    assert len(p) == 2
    assert len(v) == 2
    assert len(o) == 1
    assert len(code) == 6
    tpv = "{}{}{}{}{}{}".format(_API, test, p, v, o, code)
    assert len(tpv) == 17
    return tpv


def newMagicCode(seed=None):
    """Generate a new random magic code"

    Args:
       seed: seed for the random number generator, or ``None`` for
             something reasonable (i.e.., current time).

    Returns:
       str: the magic code
    """
    random.seed(seed)
    magic = str(random.randrange(0, 10 ** 6)).zfill(6)
    assert len(magic) == 6
    return magic



class SeriousException(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class PlomSeriousException(SeriousException):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class BenignException(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class PlomAPIException(BenignException):
    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)


class PlomBenignException(BenignException):
    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)


class PlomNoMoreException(BenignException):
    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)


class PlomAuthenticationException(BenignException):
    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)


class PlomLatexException(BenignException):
    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)


class PlomExistingLoginException(BenignException):
    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2018-2020 Andrew Rechnitzer
# Copyright (C) 2020 Dryden Wiebe
# Copyright (C) 2020 Vala Vakilian
# Copyright (C) 2020 Colin B. Macdonald

import uuid
from passlib.context import CryptContext


class Authority:
    """A class to do all our authentication - passwords and tokens."""

    def __init__(self, masterToken):
        """Set up cryptocontext, userlist and tokenlist"""
        self.ctx = CryptContext(schemes=["pbkdf2_sha256", "bcrypt"], deprecated="auto")
        # set master token, being careful of what user supplies
        # is hex string of uuid4
        self.masterToken = self.build_master_token(masterToken)
        # convert it to int - is base 16
        self.mti = int(self.masterToken, 16)
        # we need this for xor-ing.

    def build_master_token(self, token):
        """Creates a new masterToken or validates existing masterToken.

        Arguments:
            token {masterToken} -- Current masterToken, if None, a new one is created.

        Returns:
            masterToken -- Valid masterToken.
        """
        if token is None:
            masterToken = uuid.uuid4().hex
            print("No masterToken given, creating one")
        else:
            try:
                masterToken = uuid.UUID(token).hex
                print("Supplied masterToken is valid. Using that.")
            except ValueError:
                masterToken = uuid.uuid4().hex
                print(
                    "Supplied masterToken is not a valid UUID. Creating new masterToken."
                )
        return masterToken

    def get_master_token(self):
        """Getter for the masterToken"""
        return self.masterToken

    def check_password(self, password, expected_hash):
        """Check the password against expected hashed password.

        Arguments:
            password (str): password to check.
            expected_hash (str): hashed password on file or None
                if we have no such user on file.

        Returns:
            bool: True on match, False otherwise.
        """
        return self.ctx.verify(password, expected_hash)

    def create_token(self):
        """Create a token for a validated user, return that token as hex and int-xor'd version for storage.

        Returns:
            list -- Client and storage tokens in the form [clientToken, storageToken].
        """
        clientToken = uuid.uuid4().hex
        storageToken = hex(int(clientToken, 16) ^ self.mti)
        return [clientToken, storageToken]

    def validate_token(self, clientToken, storageToken):
        """Validates a given token against the storageToken.

        Arguments:
            clientToken (str): The token, a hex string provided by the
                client.  This may be untrusted unsanitized input.
            storageToken (str): The token we are checking against.

        Returns:
            bool/None: True if validated, False/None otherwise.  False
                indicates an invalid token.  None indicates a malformed
                token.  This means `if validate_token(...):` works.
        """
        if not isinstance(clientToken, str):
            return None
        # should not be significantly longer than UUID's 32 hex digits
        if len(clientToken) > 64:
            return None
        try:
            clientTokenInt = int(clientToken, 16)
        except ValueError:
            return None
        if hex(clientTokenInt ^ self.mti) == storageToken:
            return True
        return False

    def check_string_is_UUID(self, tau):
        """Checks that a given string is a valid UUID.

        Arguments:
            tau {str} -- String to check.

        Returns:
            bool -- True if a valid UUID, False otherwise.
        """
        try:
            token = uuid.UUID(tau)
        except ValueError:
            return False
        return True

    def basic_user_password_check(self, username, password):
        """Sanity check for usernames and passwords.

        Arguments:
            username {str} -- Given username.
            password {str} -- Given password.

        Returns:
            bool -- True if valid, false otherwise.
        """
        # username must be length 4 and alphanumeric
        if not (len(username) >= 4 and username.isalnum()):
            return False
        # password must be length 4 and not contain username.
        if (len(password) < 4) or (username in password):
            return False
        return True

    def create_password_hash(self, password):
        """Creates a hash of a string password.

        Arguments:
            password {str} -- Password to hash.

        Returns:
            hex -- Hashed password.
        """
        return self.ctx.hash(password)



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2020 Andrew Rechnitzer
# Copyright (C) 2020 Colin B. Macdonald
# Copyright (C) 2020 Dryden Wiebe

"""
Digit hunter extracts random digit images from the MNIST database
and stores then in a json file (used to create digits.json).  The
results are used by faketools.py to create ID-pages for the plom-demo.
This python program not run on a normal Plom (or demo) run, however it
is being kept in the repo for posterity.

Dataset used: https://en.wikipedia.org/wiki/MNIST_database
"""

__copyright__ = "Copyright (C) 2020 Andrew Rechnitzer and Colin B. Macdonald"
__credits__ = "The Plom Project Developers"
__license__ = "AGPL-3.0-or-later"

import base64
import json
from random import sample
from collections import defaultdict
import subprocess

import numpy as np
import tensorflow as tf
import cv2

# pair (x_train, y_train), (x_test, y_test) from mnist dataset
mnist = tf.keras.datasets.mnist.load_data()
names = mnist[1][1]
images = mnist[1][0]

N = 32  # how many of each digit to collect
digits = defaultdict(list)

for k in range(len(names)):
    digits[names[k]].append(k)

print("Collected random digits. Now saving them.")
imgs = []
for d in range(10):
    theDigits = sample(digits[d], N)
    for k in range(N):
        c = digits[d][k]
        img = ~images[c]  # since digit is bitwise-inverted.
        assert img.shape == (28, 28)

        # quantize
        # img = np.round(np.round(img / 255.0 * 3) / 3.0 * 255)

        # colorize
        bgr = np.zeros((28, 28, 3))
        bgr[:, :, 0] = img // 4 + 192  # blue to [192, 255]
        bgr[:, :, 1] = img
        bgr[:, :, 2] = img

        worked, buf = cv2.imencode(".png", bgr)

        # pngquant to compress to 4-colour
        # TODO: faster to use a pipe instead of so much disk access
        fname = "tmpdigit.png"
        with open(fname, "wb") as f:
            f.write(buf)
        subprocess.check_call(["pngquant", "4", "--force", fname])
        with open(fname.replace(".png", "-fs8.png"), "rb") as f:
            buf = f.read()

        # cv2.imshow("argh", img)
        if worked is False:
            print("EEK - problem")
            quit()
        imgs.append(base64.b64encode(buf).decode())

with open("digits.json", "w") as fh:
    json.dump(imgs, fh)



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2020 Victoria Schuster
# Copyright (C) 2020 Andrew Rechnitzer

from PyQt5.QtCore import (
    QTimer,
    QPropertyAnimation,
    QByteArray,
    QBuffer,
    QIODevice,
    QPoint,
    QPointF,
    pyqtProperty,
)
from PyQt5.QtGui import QBrush, QColor, QImage, QPixmap, QPen
from PyQt5.QtWidgets import (
    QUndoCommand,
    QGraphicsItem,
    QGraphicsPixmapItem,
    QGraphicsObject,
    QGraphicsSceneMouseEvent,
    QDialog,
    QSpinBox,
    QCheckBox,
    QVBoxLayout,
    QLabel,
    QDialogButtonBox,
    QGroupBox,
    QFormLayout,
)


class CommandImage(QUndoCommand):
    """ A class for making image commands. """

    def __init__(self, scene, pt, image, scale=1, border=True, data=None):
        """
        Initialize a new Image command.

        Args:
            scene (PageScene): the scene the image is being inserted into.
            pt (QPoint): the point of the top left corner of the image.
            image (QImage): the image being added to the scene.
            scale (float): the scaling value, <1 decreases size, >1 increases.
            border (bool): True if the image has a border, false otherwise.
            data (str): Base64 data held in a string if the image had
                previously been json serialized.
        """
        super(CommandImage, self).__init__()
        self.scene = scene
        self.width = image.width()
        if data is None:
            toMidpoint = QPoint(-image.width() / 2, -image.height() / 2)
            self.midPt = pt + toMidpoint
        else:
            self.midPt = pt
        self.image = image
        self.imageItem = ImageItemObject(self.midPt, self.image, scale, border, data)
        self.setText("Image")

    @classmethod
    def from_pickle(cls, X, *, scene):
        """Construct a CommandImage from a serialized form."""
        assert X[0] == "Image"
        X = X[1:]
        if len(X) != 5:
            raise ValueError("wrong length of pickle data")
        # extract data from encoding
        # TODO: sus arithmetic here
        data = QByteArray().fromBase64(bytes(X[2][2 : len(X[2]) - 2], encoding="utf-8"))
        img = QImage()
        if not img.loadFromData(data):
            log.error("Encountered a problem loading image.")
            raise ValueError("Encountered a problem loading image.")
        return cls(scene, QPointF(X[0], X[1]), img, X[3], X[4], X[2])

    def redo(self):
        """ Redoes adding the image to the scene. """
        self.imageItem.flash_redo()
        self.scene.addItem(self.imageItem.ci)

    def undo(self):
        """ Undoes adding the image to the scene. """
        self.imageItem.flash_undo()
        QTimer.singleShot(200, lambda: self.scene.removeItem(self.imageItem.ci))


class ImageItemObject(QGraphicsObject):
    """A class which encapsulates the QImage.

    Used primarily for animation when undo or redo is performed.
    """

    def __init__(self, midPt, image, scale, border, data):
        """
        Initializes an new ImageItemObject.

        Args:
            midPt (QPoint): the point middle of the image.
            image (QImage): the image being added to the scene.
            scale (float): the scaling value, <1 decreases size, >1 increases.
            border (bool): True if the image has a border, false otherwise.
            data (str): Base64 data held in a string if the image had
                previously been json serialized.
        """
        super(ImageItemObject, self).__init__()
        self.ci = ImageItem(midPt, image, self, scale, border, data)
        self.anim = QPropertyAnimation(self, b"thickness")
        self.border = border

    def flash_undo(self):
        """Animates the object in an undo sequence."""
        self.anim.setDuration(200)
        if self.ci.border:
            self.anim.setStartValue(4)
        else:
            self.anim.setStartValue(0)
        self.anim.setKeyValueAt(0.5, 8)
        self.anim.setEndValue(0)
        self.anim.start()

    def flash_redo(self):
        """Animates the object in a redo sequence. """
        self.anim.setDuration(200)
        self.anim.setStartValue(0)
        self.anim.setKeyValueAt(0.5, 8)
        if self.ci.border:
            self.anim.setEndValue(4)
        else:
            self.anim.setEndValue(0)
        self.anim.start()

    @pyqtProperty(int)
    def thickness(self):
        return self.ci.thick

    @thickness.setter
    def thickness(self, value):
        self.ci.thick = value
        self.ci.update()


class ImageItem(QGraphicsPixmapItem):
    """
    An image added to a paper.
    """

    def __init__(self, midPt, qImage, parent, scale, border, data):
        """
        Initialize a new ImageItem.

        Args:
            pt (QPoint): the point of the top left corner of the image.
            qImage (QImage): the image being added to the scene.
            scale (float): the scaling value, <1 decreases size, >1 increases.
            border (bool): True if the image has a border, false otherwise.
            data (str): Base64 data held in a string if the image had
                previously been json serialized.
        """
        super(ImageItem, self).__init__()
        self.qImage = qImage
        self.border = border
        self.setPixmap(QPixmap.fromImage(self.qImage))
        self.setPos(midPt)
        self.setFlag(QGraphicsItem.ItemIsMovable)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges)
        self.saveable = True
        self.animator = [parent]
        self.animateFlag = False
        self.parent = parent
        self.setScale(scale)
        self.data = data
        self.thick = 0

    def paint(self, painter, option, widget=None):
        """
        Paints the scene by adding a red border around the image if applicable.
        """
        if not self.scene().itemWithinBounds(self):
            # paint a bounding rectangle out-of-bounds warning
            painter.setPen(QPen(QColor(255, 165, 0), 8))
            painter.setBrush(QBrush(QColor(255, 165, 0, 128)))
            painter.drawRoundedRect(option.rect, 10, 10)

        super().paint(painter, option, widget)
        if self.thick > 0:
            painter.save()
            painter.setPen(QPen(QColor("red"), self.thick))
            painter.drawRect(self.boundingRect())
            painter.restore()

    def pickle(self):
        """
        Pickle the image into a list containing important information.

        Returns:
            (list): containing
                (str): "Image"
                (float): X position of image.
                (float): Y position of image.
                (str): a string containing image data in base64 encoding.
                (float): scale of the image.
                (bool): true if the image contains a red border,
                    false otherwise.
        """
        if self.data is None:
            ba = QByteArray()
            buffer = QBuffer(ba)
            buffer.open(QIODevice.WriteOnly)
            self.qImage.save(buffer, "PNG")
            pickle = [
                "Image",
                self.x(),
                self.y(),
                str(ba.toBase64().data()),
                self.scale(),
                self.border,
            ]
        else:
            pickle = ["Image", self.x(), self.y(), self.data, self.scale(), self.border]
        return pickle

    def mouseDoubleClickEvent(self, event: "QGraphicsSceneMouseEvent"):
        """
        On double-click, show menu and modify image according to user inputs.

        Args:
            event (QMouseEvent): the double mouse click.

        Returns:
            None
        """
        dialog = ImageSettingsDialog(int(self.scale() * 100), self.border)
        if dialog.exec():
            scale, border = dialog.getSettings()
            self.setScale(scale / 100)
            if border is not self.border:
                self.border = border
                if self.border:  # update border thickness
                    self.thick = 4
                else:
                    self.thick = 0
                self.update()  # trigger update


class ImageSettingsDialog(QDialog):
    """ Menu dialog for Image Settings. """

    NumGridRows = 2
    NumButtons = 3

    def __init__(self, scalePercent, checked):
        """
        Initialize a new image settings dialog object.

        Args:
            scalePercent (int): Scale of the image (as a percentage)
            checked (bool): True if the image currently has a red border,
                False otherwise.
        """
        super(ImageSettingsDialog, self).__init__()
        self.createFormGroupBox(scalePercent, checked)
        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)
        mainLayout = QVBoxLayout()
        mainLayout.addWidget(self.formGroupBox)
        mainLayout.addWidget(buttonBox)
        self.setLayout(mainLayout)
        self.setWindowTitle("Image Settings")

    def createFormGroupBox(self, scalePercent, checked):
        """
        Build the form Box.

        Args:
            scalePercent (int): Scale of the image (as a percentage)
            checked (bool): True if the image currently has a red border,
                False otherwise.
        Returns:
            None

        """
        self.formGroupBox = QGroupBox("Image Settings")
        layout = QFormLayout()
        self.scaleButton = QSpinBox()
        self.scaleButton.setRange(1, 500)
        self.scaleButton.setValue(scalePercent)
        layout.addRow(QLabel("Scale"), self.scaleButton)
        self.checkBox = QCheckBox()
        self.checkBox.setChecked(checked)
        layout.addRow(QLabel("Include Red Border"), self.checkBox)
        self.formGroupBox.setLayout(layout)

    def getSettings(self):
        """
        Return the settings held in the dialog box.

        Notes:
            Even if the user presses Cancel, the values will still be held
            by the dialog box. Make sure to ensure exec() returns true
            before accessing these values.
        Returns:
            (int): the scale of the image that the user has chosen.
            (bool): True if user wants image to have a red border,
                False otherwise.
        """
        return self.scaleButton.value(), self.checkBox.isChecked()



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2019-2020 Andrew Rechnitzer
# Copyright (C) 2020 Colin B. Macdonald
# Copyright (C) 2020 Vala Vakilian

from datetime import datetime
import hashlib
import imghdr
import json
import os
import subprocess
import tempfile
import uuid
import logging

from plom.textools import texFragmentToPNG


log = logging.getLogger("server")


def MgetQuestionMax(self, question_number, version_number):
    """Return the maximum score for the question.

    Args:
        question_number (int): Question number.
        version_number (int): Version number.

    Returns:
        list: A list where the first element is a boolean operation
            status response. The second element is either a string
            indicating if question fo version number is incorrect, or,
            the maximum score for this question as an integer.
    """

    version_number = int(version_number)
    question_number = int(question_number)
    # check question /version in range.
    if question_number < 1 or question_number > self.testSpec["numberOfQuestions"]:
        return [False, "QE"]
    if version_number < 1 or version_number > self.testSpec["numberOfVersions"]:
        return [False, "VE"]
    # Send back the max-mark for that question_number/version_number
    return [True, self.testSpec["question"][str(question_number)]["mark"]]


def MgetAllMax(self):
    """Get the maximum mark for each question in the exam.

    Returns:
        dict: A dictionary of the form {key: question_number, value: question_max_grade}
            for the exam questions.
    """

    max_grades = {}
    for q_number in range(1, self.testSpec["numberOfQuestions"] + 1):
        max_grades[q_number] = self.testSpec["question"][str(q_number)]["mark"]
    return max_grades


def MprogressCount(self, question_number, version_number):
    """Send back the marking progress count for question.

    Args:
        question_number (int): Question number.
        version_number (int): Version number.

    Returns:
        list: A list of two integers indicating the number of questions graded
            and the total number of assigned question to be graded of this question number
            and question version.
    """

    version_number = int(version_number)
    question_number = int(question_number)
    return [
        self.DB.McountMarked(question_number, version_number),
        self.DB.McountAll(question_number, version_number),
    ]


def MgetDoneTasks(self, username, question_number, version_number):
    """Respond with a list of the graded tasks.

    Args:
        username (str): Username string.
        question_number (int): Question number.
        version_number (int): Version number.

    Returns:
        list: Respond with a list of done tasks, each list includes
            [question_code string, maximum_mark, question_grade, question_tag string].
    """

    version_number = int(version_number)
    question_number = int(question_number)
    return self.DB.MgetDoneTasks(username, question_number, version_number)


def MgetNextTask(self, question_number, version_number):
    """Retrieve the next unmarked paper from the database.

    Args:
        question_number (int): Next question's question number.
        version_number (int): Next question's version number.

    Returns:
        list: Respond with a list with either value False or the value
            of True with the question code string for next task.
    """

    give = self.DB.MgetNextTask(question_number, version_number)
    if give is None:
        return [False]
    else:
        return [True, give]


def MlatexFragment(self, username, latex_fragment):
    """Respond with a path to the latex fragment image.

    Args:
        username (str): Username string.
        latex_fragment (str): The latex string for the latex image requested.

    Returns:
        list: A list with either False or True with the latex image's
            file name.
    """

    # TODO - only one frag file per user - is this okay?
    filename = os.path.join(self.tempDirectory.name, "{}_frag.png".format(username))

    if texFragmentToPNG(latex_fragment, filename):
        return [True, filename]
    else:
        return [False]


def MclaimThisTask(self, username, task_code):
    """Assign the specified paper to this user and return the task information.

    Args:
        username (str): User who requests the paper.
        task_code (str): Code string for the claimed task.

    Returns:
        list: A list which either only has a False value included or
            [True, `question_tag`, `integrity_check`, `list_of_image_md5s` `image_file1`, `image_file2`,...]
    """

    return self.DB.MgiveTaskToClient(username, task_code)


def MdidNotFinish(self, username, task_code):
    """Inform database that a user did not finish a task.

    Args:
        username (str): Owner of the unfinished task.
        task_code (str): Code string for the unfinished task.
    """

    self.DB.MdidNotFinish(username, task_code)
    return


# TODO: As input to MreturnMarkedTask, the comments string is in a
# list ie `["comment 1", "comment 2", "comment 3"]`
# Maybe this should be changed.
def MreturnMarkedTask(
    self,
    username,
    task_code,
    question_number,
    version_number,
    mark,
    image,
    plomdat,
    comments,
    time_spent_marking,
    tags,
    md5_code,
    integrity_check,
    image_md5s,
):
    """Save the marked paper's information to database and respond with grading progress.

    Args:
        username (str): User who marked the paper.
        task_code (str): Code string for the task.
        question_number (int): Marked queston number.
        version_number (int): Marked question version number.
        mark (int): Question mark.
        image (bytearray): Marked image of question.
        plomdat (bytearray): Plom data file used for saving marking information in
            editable format.
        comments (str): Return the String of the comments list.
        time_spent_marking (int): Seconds spent marking the paper.
        tags (str): Tag assigned to the paper.
        md5_code (str): MD5 hash key for this task.
        integrity_check (str): the integrity_check string for this task
        image_md5s (list[str]): list of image md5sums used.


    Returns:
        list: Respond with a list which includes:
            [False, Error message of either mismatching codes or database owning the task.]
            [True, number of graded tasks, total number of tasks.]
    """

    # TODO: score + file sanity checks were done at client. Do we need to redo here?
    # image, plomdat are bytearrays, comments = list
    annotated_filename = "markedQuestions/G{}.png".format(task_code[1:])
    plom_filename = "markedQuestions/plomFiles/G{}.plom".format(task_code[1:])
    comments_filename = "markedQuestions/commentFiles/G{}.json".format(task_code[1:])

    # do sanity checks on incoming annotation image file
    # Check the annotated_filename is valid png - just check header presently
    # notice that 'imghdr.what(name, h=blah)' ignores the name, instead checks stream blah.
    if imghdr.what(annotated_filename, h=image) != "png":
        log.error(
            "Uploaded annotation file is not a PNG. Instead is = {}".format(
                imghdr.what(annotated_filename, h=image)
            )
        )
        return [False, "Misformed image file. Try again."]

    # Also check the md5sum matches
    md5n = hashlib.md5(image).hexdigest()
    if md5_code != md5n:
        log.error(
            "Mismatched between client ({}) and server ({}) md5sums of annotated image.".format(
                md5_code, md5n
            )
        )
        return [
            False,
            "Misformed image file - md5sum doesn't match serverside={} vs clientside={}. Try again.".format(
                md5n, md5_code
            ),
        ]

    # now update the database
    database_task_response = self.DB.MtakeTaskFromClient(
        task_code,
        username,
        mark,
        annotated_filename,
        plom_filename,
        comments_filename,
        time_spent_marking,
        tags,
        md5n,
        integrity_check,
        image_md5s,
    )

    if database_task_response[0] is False:
        return database_task_response

    # db successfully updated
    #  check if those files exist already - back up if so
    for filename in [annotated_filename, plom_filename, comments_filename]:
        if os.path.isfile(filename):
            os.rename(
                filename,
                filename + ".rgd" + datetime.now().strftime("%d_%H-%M-%S"),
            )

    # now write in the files
    with open(annotated_filename, "wb") as file_header:
        file_header.write(image)
    with open(plom_filename, "wb") as file_header:
        file_header.write(plomdat)
    with open(comments_filename, "w") as file_header:
        json.dump(comments, file_header)

    self.MrecordMark(username, mark, annotated_filename, time_spent_marking, tags)
    # return ack with current counts.
    return [
        True,
        self.DB.McountMarked(question_number, version_number),
        self.DB.McountAll(question_number, version_number),
    ]


def MrecordMark(self, username, mark, annotated_filename, time_spent_marking, tags):
    """Saves the marked paper information as a backup, independent of the server

    Args:
        username (str): User who marked the paper.
        mark (int): Question mark.
        annotated_filename (str): Name of the annotated image file.
        time_spent_marking (int): Seconds spent marking the paper.
        tags (str): Tag assigned to the paper.
    """

    with open("{}.txt".format(annotated_filename), "w") as file_header:
        file_header.write(
            "{}\t{}\t{}\t{}\t{}\t{}".format(
                annotated_filename,
                mark,
                username,
                datetime.now().strftime("%Y-%m-%d,%H:%M"),
                time_spent_marking,
                tags,
            )
        )


def MgetImages(self, username, task_code, integrity_check):
    """Respond with paths to the marked and original images of a marked question.

    Args:
        username (str): User who marked the paper.
        task_code (str): Code string for the task.
        integrity_check (str): Integrity check string for the task.

    Returns:
        list: A list of the format:
            [False, Error message string.]
            [True, Number of papers in the question, md5 list, Original images paths,
            Annotated image path, Plom data file for this page]
    """

    return self.DB.MgetImages(username, task_code, integrity_check)


# TODO: Have to figure this out.  Please needs documentation.
def MgetOriginalImages(self, task):
    return self.DB.MgetOriginalImages(task)


def MsetTag(self, username, task_code, tag):
    """Assign a tag string to a paper

    Args:
        username (str): User who assigned tag to the paper.
        task_code (str): Code string for the task.
        tags (str): Tag assigned to the paper.

    Returns:
        bool: True or False indicating if tag was set in database successfully.
    """

    return self.DB.MsetTag(username, task_code, tag)


def MgetWholePaper(self, test_number, question_number):
    """Respond with all the images of the paper including the given question.

    Args:
        test_number (str): A string which has the test number in the format `0011`
            for example.
        question_number (str): Question number.

    Returns:
        list: A list including the following information:
            Boolean of wether we got the paper images.
            A list of lists including [`test_version`, `image_md5sum_list`, `does_page_belong_to_question`].
            Followed by a series of image paths for the pages of the paper.
    """

    return self.DB.MgetWholePaper(test_number, question_number)


def MreviewQuestion(self, test_number, question_number, version_number):
    """Save question results in database.

    Args:
        test_number (int): Reviewed test number.
        question_number (int): Review queston number.
        version_number (int): Marked question version number.

    Returns:
        list: Respond with a list with value of either True or False.
    """

    return self.DB.MreviewQuestion(test_number, question_number, version_number)


# TODO: Deprecated.
# TODO: Should be removed.
def MrevertTask(self, code):
    rval = self.DB.MrevertTask(code)
    # response is [False, "NST"] or [False, "NAC"] or [True]
    return rval



from .misc_utils import format_int_list_with_runs


def test_runs():
    L = ["1", "2", "3", "4", "7", "10", "11", "12", "13", "14", "64"]
    uout = "1–4, 7, 10–14, 64"
    aout = "1-4, 7, 10-14, 64"
    assert format_int_list_with_runs(L, use_unicode=True) == uout
    assert format_int_list_with_runs(L, use_unicode=False) == aout
    assert format_int_list_with_runs(L) in (aout, uout)


def test_shortruns():
    L = ["1", "2", "4", "5", "6", "7", "9", "10", "12", "78", "79", "80"]
    out = "1, 2, 4-7, 9, 10, 12, 78-80"
    assert format_int_list_with_runs(L, use_unicode=False) == out



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2018-2020 Andrew Rechnitzer
# Copyright (C) 2019-2020 Colin B. Macdonald
# Copyright (C) 2020 Dryden Wiebe
# Copyright (C) 2020 Vala Vakilian

__copyright__ = "Copyright (C) 2019-2020 Andrew Rechnitzer and others"
__credits__ = ["Andrew Rechnitzer", "Colin Macdonald", "Dryden Wiebe", "Vala Vakilian"]
__license__ = "AGPLv3"

# TODO - directory structure!

import hashlib
import toml
import json
import os
import ssl
import subprocess
import sys
import tempfile
import uuid
import logging
from pathlib import Path

from aiohttp import web

from plom import __version__
from plom import Plom_API_Version as serverAPI
from plom import Default_Port
from plom import SpecParser
from plom import specdir
from plom.db import PlomDB

from .authenticate import Authority

serverInfo = {"server": "127.0.0.1", "port": Default_Port}
# ----------------------
sslContext = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH)
sslContext.check_hostname = False
sslContext.load_cert_chain(
    "serverConfiguration/plom-selfsigned.crt", "serverConfiguration/plom.key"
)


from .plomServer.routesUserInit import UserInitHandler
from .plomServer.routesUpload import UploadHandler
from .plomServer.routesID import IDHandler
from .plomServer.routesMark import MarkHandler
from .plomServer.routesReport import ReportHandler


# 5 is to keep debug/info lined up
logging.basicConfig(
    format="%(asctime)s %(levelname)5s:%(name)s\t%(message)s",
    datefmt="%b%d %H:%M:%S",
)
log = logging.getLogger("server")
# We will reset this later after we read the config
logging.getLogger().setLevel("Debug".upper())


# ----------------------
def build_directories():
    """Build the directories that this script needs"""

    # the list of directories. Might need updating.
    lst = [
        "pages",
        "pages/discardedPages",
        "pages/collidingPages",
        "pages/unknownPages",
        "pages/originalPages",
        "markedQuestions",
        "markedQuestions/plomFiles",
        "markedQuestions/commentFiles",
    ]
    for dir in lst:
        try:
            os.mkdir(dir)
            log.debug("Building directory {}".format(dir))
        except FileExistsError:
            pass


# ----------------------


class Server(object):
    def __init__(self, spec, db, masterToken):
        log.debug("Initialising server")
        self.testSpec = spec
        self.authority = Authority(masterToken)
        self.DB = db
        self.API = serverAPI
        self.Version = __version__
        print(
            "Server launching with masterToken = '{}' {}".format(
                self.authority.get_master_token(),
                type(self.authority.get_master_token()),
            )
        )
        self.tempDirectory = tempfile.TemporaryDirectory()
        # Give directory correct permissions.
        subprocess.check_call(["chmod", "o-r", self.tempDirectory.name])
        self.load_users()

    def load_users(self):
        """Load the users from json file, add them to the database and checks pwd hashes.

        It does simple sanity checks of pwd hashes to see if they have changed.
        """

        if os.path.exists("serverConfiguration/userList.json"):
            with open("serverConfiguration/userList.json") as data_file:
                # load list of users + pwd hashes
                userList = json.load(data_file)
                # for each name check if in DB by asking for the hash of its pwd
                for uname in userList.keys():
                    passwordHash = self.DB.getUserPasswordHash(uname)
                    if passwordHash is None:  # not in list
                        self.DB.createUser(uname, userList[uname])
                    else:
                        if passwordHash != userList[uname]:
                            log.warning("User {} password has changed.".format(uname))
                        self.DB.setUserPasswordHash(userList[uname], passwordHash)
            log.debug("Loading users")
        else:
            # Cannot find users - give error and quit out.
            log.error("Cannot find user/password file - aborting.")
            quit()

    from .plomServer.serverUserInit import (
        validate,
        checkPassword,
        checkUserEnabled,
        createModifyUser,
        InfoShortName,
        info_spec,
        reloadUsers,
        giveUserToken,
        setUserEnable,
        closeUser,
    )
    from .plomServer.serverUpload import (
        doesBundleExist,
        createNewBundle,
        sidToTest,
        addTestPage,
        addHWPage,
        addLPage,
        processHWUploads,
        processTUploads,
        processLUploads,
        addUnknownPage,
        addCollidingPage,
        replaceMissingTestPage,
        replaceMissingHWQuestion,
        removeAllScannedPages,
        getUnknownPageNames,
        getDiscardNames,
        getCollidingPageNames,
        getUnknownImage,
        getCollidingImage,
        getDiscardImage,
        getTPageImage,
        getHWPageImage,
        getEXPageImage,
        getLPageImage,
        getQuestionImages,
        getAllTestImages,
        checkTPage,
        removeUnknownImage,
        removeCollidingImage,
        unknownToTestPage,
        unknownToHWPage,
        unknownToExtraPage,
        collidingToTestPage,
        discardToUnknown,
    )
    from .plomServer.serverID import (
        IDprogressCount,
        IDgetNextTask,
        IDgetDoneTasks,
        IDgetImage,
        IDgetImageFromATest,
        IDclaimThisTask,
        IDdidNotFinish,
        id_paper,
        ID_id_paper,
        IDdeletePredictions,
        IDrunPredictions,
        IDreviewID,
    )
    from .plomServer.serverMark import (
        MgetAllMax,
        MprogressCount,
        MgetQuestionMax,
        MgetDoneTasks,
        MgetNextTask,
        MlatexFragment,
        MclaimThisTask,
        MdidNotFinish,
        MrecordMark,
        MreturnMarkedTask,
        MgetImages,
        MgetOriginalImages,
        MsetTag,
        MgetWholePaper,
        MreviewQuestion,
        MrevertTask,
    )

    from .plomServer.serverReport import (
        RgetUnusedTests,
        RgetScannedTests,
        RgetIncompleteTests,
        RgetCompleteHW,
        RgetMissingHWQ,
        RgetProgress,
        RgetQuestionUserProgress,
        RgetMarkHistogram,
        RgetMarked,
        RgetIdentified,
        RgetCompletionStatus,
        RgetOutToDo,
        RgetStatus,
        RgetSpreadsheet,
        RgetCoverPageInfo,
        RgetOriginalFiles,
        RgetAnnotatedFiles,
        RgetMarkReview,
        RgetIDReview,
        RgetAnnotatedImage,
        RgetUserList,
        RgetUserDetails,
    )


def get_server_info():
    """Read the server info from config file."""

    global serverInfo
    try:
        with open("serverConfiguration/serverDetails.toml") as data_file:
            serverInfo = toml.load(data_file)
            logging.getLogger().setLevel(serverInfo["LogLevel"].upper())
            log.debug("Server details loaded: {}".format(serverInfo))
    except FileNotFoundError:
        log.warning("Cannot find server details, using defaults")
    # Special treatment for chatty modules
    # TODO: nicer way to do this?
    if serverInfo["LogLevel"].upper() == "INFO":
        logging.getLogger("aiohttp.access").setLevel("WARNING")


def launch(masterToken=None):
    """Launches the Plom server.

    Keyword Arguments:
        masterToken {str} -- Token that is authenticated by the authority, if None, one is created in authenticate.py. default: {None}
    """

    log.info("Plom Server {} (communicates with api {})".format(__version__, serverAPI))
    get_server_info()
    examDB = PlomDB(Path(specdir) / "plom.db")
    spec = SpecParser(Path(specdir) / "verifiedSpec.toml").spec
    build_directories()
    peon = Server(spec, examDB, masterToken)
    userIniter = UserInitHandler(peon)
    uploader = UploadHandler(peon)
    ider = IDHandler(peon)
    marker = MarkHandler(peon)
    reporter = ReportHandler(peon)

    try:
        # construct the web server
        app = web.Application()
        # add the routes
        log.info("Setting up routes")
        userIniter.setUpRoutes(app.router)
        uploader.setUpRoutes(app.router)
        ider.setUpRoutes(app.router)
        marker.setUpRoutes(app.router)
        reporter.setUpRoutes(app.router)
        # run the web server
        log.info("Start the server!")
        web.run_app(app, ssl_context=sslContext, port=serverInfo["port"])
    except KeyboardInterrupt:
        log.info("Closing down")  # TODO: I never see this!
        pass



from plom.db.tables import *
from datetime import datetime

import logging

log = logging.getLogger("DB")

# ------------------
# Marker stuff


def McountAll(self, q, v):
    """Count all the scanned q/v groups."""
    try:
        return (
            QGroup.select()
            .join(Group)
            .where(
                QGroup.question == q,
                QGroup.version == v,
                Group.scanned == True,
            )
            .count()
        )
    except QGroup.DoesNotExist:
        return 0


def McountMarked(self, q, v):
    """Count all the q/v groups that have been marked."""
    try:
        return (
            QGroup.select()
            .join(Group)
            .where(
                QGroup.question == q,
                QGroup.version == v,
                QGroup.status == "done",
                Group.scanned == True,
            )
            .count()
        )
    except QGroup.DoesNotExist:
        return 0


def MgetDoneTasks(self, user_name, q, v):
    """When a marker-client logs on they request a list of papers they have already marked.
    Send back the list of [group-ids, mark, marking_time, tags] for each paper.
    """
    uref = User.get(name=user_name)  # authenticated, so not-None

    query = QGroup.select().where(
        QGroup.user == uref,
        QGroup.question == q,
        QGroup.version == v,
        QGroup.status == "done",
    )
    mark_list = []
    for qref in query:  # grab that questionData object
        aref = qref.annotations[-1]  # grab the last annotation
        mark_list.append(
            [
                qref.group.gid,
                aref.mark,
                aref.marking_time,
                aref.tags,
                aref.integrity_check,
            ]
        )
        # note - used to return qref.status, but is redundant since these all "done"
    log.debug('Sending completed Q{}v{} tasks to user "{}"'.format(q, v, user_name))
    return mark_list


def MgetNextTask(self, q, v):
    """Find unmarked (but scanned) q/v-group and send the group-id back to client."""
    with plomdb.atomic():
        try:
            qref = (
                QGroup.select()
                .join(Group)
                .where(
                    QGroup.status == "todo",
                    QGroup.question == q,
                    QGroup.version == v,
                    Group.scanned == True,
                )
                .get()
            )
        except QGroup.DoesNotExist as e:
            log.info("Nothing left on Q{}v{} to-do pile".format(q, v))
            return None

        log.debug("Next Q{}v{} task = {}".format(q, v, qref.group.gid))
        return qref.group.gid


def MgiveTaskToClient(self, user_name, group_id):
    """Assign a marking task to a certain user, and give them back needed data.

    args:
        user_name (str): the user name who is claiming the task.
        group_id (TODO): somehow tells the task (?).

    Return:
        list: `[False]` on error.  TODO: different cases handled?  Issue #1267.
            Otherwise, the list is
                `[True, metadata, tags, integrity_check]`
            where each row of `metadata` consists of
                `[DB_id, md5_sum, server_filename]`
            Note: `server_filename` is implementation-dependent, could change
            without notice, etc.  Clients could use this to get hints for a
            a local file name for example.

    question/version via group_id as a task to the given user, unless has been
    taken by another user.

    Create new annotation by copying the last one for that qdata - pages created when returned.
    """

    uref = User.get(name=user_name)  # authenticated, so not-None

    with plomdb.atomic():
        gref = Group.get_or_none(Group.gid == group_id)
        if gref is None:  # this should not happen.
            log.info("That question {} not known".format(group_id))
            return [False]
        if gref.scanned == False:  # this should not happen either
            log.info("That question {} not scanned".format(group_id))
            return [False]
        # grab the qdata corresponding to that group
        qref = gref.qgroups[0]
        if (qref.user is not None) and (
            qref.user != uref
        ):  # has been claimed by someone else.
            return [False]
        # update status, username
        qref.status = "out"
        qref.user = uref
        qref.time = datetime.now()
        qref.save()
        # we give the marker the pages from the **existing** annotation
        # (when task comes back we create the new pages, new annotation etc)
        if len(qref.annotations) < 1:
            log.error(
                "unexpectedly, len(aref.annotations) = {}".format(len(qref.annotations))
            )
            log.error("qref={}, group_id={}".format(qref, group_id))
            return [False]
        aref = qref.annotations[-1]  # are these in right order (TODO?)
        image_metadata = []
        for p in aref.apages.order_by(APage.order):
            image_metadata.append([p.image.id, p.image.md5sum, p.image.file_name])
        # update user activity
        uref.last_action = "Took M task {}".format(group_id)
        uref.last_activity = datetime.now()
        uref.save()
        log.debug(
            'Giving marking task {} to user "{}" with integrity_check {}'.format(
                group_id, user_name, aref.integrity_check
            )
        )
        return [True, image_metadata, aref.tags, aref.integrity_check]


def MdidNotFinish(self, user_name, group_id):
    """When user logs off, any images they have still out should be put
    back on todo pile. This returns the given gid to the todo pile.
    """
    uref = User.get(name=user_name)  # authenticated, so not-None

    with plomdb.atomic():
        gref = Group.get_or_none(Group.gid == group_id)
        if gref is None:  # this should not happen.
            log.info("That task {} not known".format(group_id))
            return
        if gref.scanned == False:  # sanity check
            return  # should not happen
        qref = gref.qgroups[0]
        # sanity check that user has task
        if qref.user != uref or qref.status != "out":
            return  # has been claimed by someone else. Should not happen

        # update status, etc
        qref.status = "todo"
        qref.user = None
        qref.marked = False
        qref.time = datetime.now()
        # now clean up the qgroup
        qref.test.marked = False
        qref.test.save()
        qref.save()
        # Log user returning given task.
        log.info("User {} did not mark task {}".format(user_name, group_id))


def MgetOneImageFilename(self, user_name, task, image_id, md5):
    """Get the filename of one image.

    Args:
        TODO: drop user_name and task?

    Returns:
        list: [True, file_name] or [False, error_msg] where
            error_msg is `"no such image"` or `"wrong md5sum"`.
            file_name is a string.
    """
    with plomdb.atomic():
        iref = Image.get_or_none(id=image_id)
        if iref is None:
            log.warning(
                "User {} asked for a non-existent image with id={}".format(
                    user_name, image_id
                )
            )
            return [False, "no such image"]
        if iref.md5sum != md5:
            log.warning(
                "User {} asked for image id={} but supplied wrong md5sum".format(
                    user_name, image_id
                )
            )
            return [False, "wrong md5sum"]
        return [True, iref.file_name]


def MtakeTaskFromClient(
    self,
    task,
    user_name,
    mark,
    annot_fname,
    plom_fname,
    comment_fname,
    marking_time,
    tags,
    md5,
    integrity_check,
    image_md5_list,
):
    """Get marked image back from client and update the record
    in the database.
    Update the annotation.
    Check to see if all questions for that test are marked and if so update the test's 'marked' flag.
    """
    uref = User.get(name=user_name)  # authenticated, so not-None

    with plomdb.atomic():
        # make sure all returned image-ids are actually images
        # keep the refs for apage creation
        image_ref_list = []
        for img_md5 in image_md5_list:
            iref = Image.get_or_none(md5sum=img_md5)
            if iref:
                image_ref_list.append(iref)
            else:
                return [False, "No_such_image"]

        # grab the group corresponding to that task
        gref = Group.get_or_none(Group.gid == task)
        if gref is None or gref.scanned is False:  # this should not happen
            log.warning(
                "That returning marking task number {} / user {} pair not known".format(
                    task, user_name
                )
            )
            return [False, "no_such_task"]
        # and grab the qdata of that group
        qref = gref.qgroups[0]
        if qref.user != uref:  # this should not happen
            return [False, "not_owner"]  # has been claimed by someone else.
        # check the integrity_check code against the db
        # TODO: suspicious: client should probably tell us what annotation its work was based-on...
        oldaref = qref.annotations[-1]
        if oldaref.integrity_check != integrity_check:
            return [False, "integrity_fail"]
        # check all the images actually come from this test - sanity check against client error
        tref = qref.test
        # make a list of all the image-md5sums of all pages assoc with tref
        test_image_md5s = []
        for pref in tref.tpages:
            # tpages are always present - whether or not anything uploaded
            if pref.scanned:  # so only check scanned ones
                test_image_md5s.append(pref.image.md5sum)
        # other page types only present if used
        for pref in tref.hwpages:
            test_image_md5s.append(pref.image.md5sum)
        for pref in tref.expages:
            test_image_md5s.append(pref.image.md5sum)
        for pref in tref.lpages:
            test_image_md5s.append(pref.image.md5sum)
        # check image_id_list against this list
        for img_md5 in image_md5_list:
            if img_md5 not in test_image_md5s:
                return [False, "image_not_in_test"]

        aref = Annotation.create(
            qgroup=qref,
            user=uref,
            edition=oldaref.edition + 1,
            tags=tags,
            time=datetime.now(),
            integrity_check=oldaref.integrity_check,
        )
        # create apages from the image_ref_list.
        ord = 0
        for iref in image_ref_list:
            ord += 1
            APage.create(annotation=aref, order=ord, image=iref)

        # update status, mark, annotate-file-name, time, and
        # time spent marking the image
        qref.status = "done"
        qref.time = datetime.now()
        qref.marked = True
        # the bundle for this image is given by the (fixed) bundle for the parent qgroup.
        aref.aimage = AImage.create(file_name=annot_fname, md5sum=md5)
        aref.mark = mark
        aref.plom_file = plom_fname
        aref.comment_file = comment_fname
        aref.marking_time = marking_time
        qref.save()
        aref.save()
        # update user activity
        uref.last_action = "Returned M task {}".format(task)
        uref.last_activity = datetime.now()
        uref.save()
        # since this has been marked - check if all questions for test have been marked
        log.info(
            "Task {} marked {} by user {} and placed at {} with md5 = {}".format(
                task, mark, user_name, annot_fname, md5
            )
        )
        # check if there are any unmarked questions left in the test
        if QGroup.get_or_none(QGroup.test == tref, QGroup.marked == False) is not None:
            log.info("Still unmarked questions in test {}".format(tref.test_number))
            return [True, "more"]

        tref.marked = True
        tref.save()
        return [True, "test_done"]


def MgetImages(self, user_name, task, integrity_check):
    """Send image list back to user for the given marking task.
    If question has been annotated then send back the annotated image and the plom file as well.
    Use integrity_check to make sure client is not asking for something outdated.

    Returns:
        list: On error, return `[False, msg]`, maybe details in 3rd entry.
            On success it can be:
            `[True, metadata]`
            Or if annotated already:
            `[True, metadata, annotatedFile, plom_file]`

    """
    uref = User.get(name=user_name)  # authenticated, so not-None
    with plomdb.atomic():
        gref = Group.get_or_none(Group.gid == task)
        # some sanity checks
        if gref is None:
            log.info("Mgetimage - task {} not known".format(task))
            return [False, "no_such_task"]
        if gref.scanned == False:  # this should not happen either
            return [False, "no_such_task"]
        # grab associated qdata
        qref = gref.qgroups[0]
        if qref.user != uref:
            # belongs to another user - should not happen
            return [
                False,
                "owner",
                "Task {} does not belong to user {}".format(task, user_name),
            ]
        # check the integrity_check code against the db
        aref = qref.annotations[-1]
        if aref.integrity_check != integrity_check:
            return [False, "integrity_fail"]
        metadata = []
        for p in aref.apages.order_by(APage.order):
            metadata.append([p.image.id, p.image.md5sum, p.image.file_name])
        rval = [True, metadata]
        if aref.aimage is not None:
            rval.extend([aref.aimage.file_name, aref.plom_file])
        return rval


def MgetOriginalImages(self, task):
    """Return the original (unannotated) page images of the given task to the user.

    Differs from MgetImages in that you need not be the owner.
    """
    with plomdb.atomic():
        gref = Group.get_or_none(Group.gid == task)
        if gref is None:  # should not happen
            log.info("MgetOriginalImages - task {} not known".format(task))
            return [False, "Task {} not known".format(task)]
        if gref.scanned == False:
            log.warning(
                "MgetOriginalImages - task {} not completely scanned".format(task)
            )
            return [False, "Task {} is not completely scanned".format(task)]
        aref = gref.qgroups[0].annotations[0]  # the original annotation pages
        # return [true, page1,..,page.n]
        rval = [True]
        for p in aref.apages.order_by(APage.order):
            rval.append(p.image.file_name)
        return rval


def MsetTag(self, user_name, task, tag):
    """Set tag on last annotation of given task.

    TODO: scary that its the last annotation: maybe client should be telling us which one?
    """

    uref = User.get(name=user_name)  # authenticated, so not-None
    with plomdb.atomic():
        gref = Group.get_or_none(Group.gid == task)
        if gref is None:  # should not happen
            log.error("MsetTag -  task {} not known".format(task))
            return False
        qref = gref.qgroups[0]
        if qref.user != uref:
            return False  # not your task - should not happen
        # grab the last annotation
        aref = qref.annotations[-1]
        if aref.user != uref:
            return False  # not your annotation - should not happen
        # update tag
        aref.tags = tag
        aref.save()
        log.info('Task {} tagged by user "{}": "{}"'.format(task, user_name, tag))
        return True


def MgetWholePaper(self, test_number, question):
    """Send page images of whole paper back to user, highlighting which belong to the given question. Do not show ID pages."""

    # we can show show not totally scanned test.
    tref = Test.get_or_none(Test.test_number == test_number)
    if tref is None:  # don't know that test - this shouldn't happen
        return [False]
    pageData = []  # for each page append a 4-tuple [
    # page-code = t.pageNumber, h.questionNumber.order, 3.questionNumber.order, or l.order
    # image-md5sum,
    # true/false - if belongs to the given question or not.
    # position in current annotation (or none if not)
    pageFiles = []  # the corresponding filenames.
    question = int(question)
    # get the current annotation and position of images within it.
    qref = QGroup.get_or_none(test=tref, question=question)
    if qref is None:  # this should not happen
        return [False]
    # dict of image-ids and positions in the current annotation
    current_image_orders = {}
    aref = qref.annotations[-1]
    if aref.apages.count() == 0:
        # this should never happen (?) no such thing as a "fresh annotation" any more
        log.critical("Oh my, colin thought it cannot happen aref={}".format(aref))
        raise RuntimeError("Oh my, colin thought it cannot happen")
        # return [False]
    for pref in aref.apages:
        current_image_orders[pref.image.id] = pref.order
    # give TPages (aside from ID pages), then HWPages, then EXPages, and then LPages
    for p in tref.tpages.order_by(TPage.page_number):
        if p.scanned is False:  # skip unscanned testpages
            continue
        if p.group.group_type == "i":  # skip IDpages (but we'll include dnm pages)
            continue
        val = [
            "t{}".format(p.page_number),
            p.image.md5sum,
            False,
            current_image_orders.get(p.image.id),
            p.image.id,
        ]
        # check if page belongs to our question
        if p.group.group_type == "q" and p.group.qgroups[0].question == question:
            val[2] = True
        pageData.append(val)
        pageFiles.append(p.image.file_name)
    # give HW and EX pages by question
    for qref in tref.qgroups.order_by(QGroup.question):
        for p in qref.group.hwpages:
            val = [
                "h{}.{}".format(qref.question, p.order),
                p.image.md5sum,
                False,
                current_image_orders.get(p.image.id),
                p.image.id,
            ]
            if qref.question == question:  # check if page belongs to our question
                val[2] = True
            pageData.append(val)
            pageFiles.append(p.image.file_name)
        for p in qref.group.expages:
            val = [
                "e{}.{}".format(qref.question, p.order),
                p.image.md5sum,
                False,
                current_image_orders.get(p.image.id),
                p.image.id,
            ]
            if qref.question == question:  # check if page belongs to our question
                val[2] = True
            pageData.append(val)
            pageFiles.append(p.image.file_name)
    # then give LPages
    for p in tref.lpages.order_by(LPage.order):
        pageData.append(
            [
                "l{}".format(p.order),
                p.image.md5sum,
                False,
                current_image_orders.get(p.image.id),
                p.image.id,
            ]
        )
        pageFiles.append(p.image.file_name)
    return [True, pageData] + pageFiles


def MreviewQuestion(self, test_number, question, version):
    """Give ownership of the given marking task to the reviewer."""
    # shift ownership to "reviewer"
    revref = User.get(name="reviewer")  # should always be there

    tref = Test.get_or_none(Test.test_number == test_number)
    if tref is None:
        return [False]
    qref = QGroup.get_or_none(
        QGroup.test == tref,
        QGroup.question == question,
        QGroup.version == version,
        QGroup.marked == True,
    )
    if qref is None:
        return [False]
    with plomdb.atomic():
        qref.user = revref
        qref.time = datetime.now()
        qref.save()
    log.info("Setting tqv {}.{}.{} for reviewer".format(test_number, question, version))
    return [True]


def MrevertTask(self, task):
    """This needs work. The qgroup is set back to its original state, the annotations (and images) are deleted, and the corresponding to-delete-filenames are returned to the server which does the actual deleting of files. In future we should probably not delete any files and just move the references within the system?"""
    gref = Group.get_or_none(Group.gid == task)
    if gref is None:
        return [False, "NST"]  # no such task
    # from the group get the test and question - all need cleaning.
    qref = gref.qgroups[0]
    tref = gref.test
    # check task is "done"
    if qref.status != "done" or qref.marked is False:
        return [False, "NAC"]  # nothing to do here
    # now update things
    log.info("Manager reverting task {}".format(task))
    with plomdb.atomic():
        # clean up test
        tref.marked = False
        tref.save()
        # clean up the qgroup
        qref.marked = False
        qref.status = "todo"
        qref.time = datetime.now()
        qref.user = None
        qref.save()
        rval = [True]  # keep list of files to delete.
        # now move existing annotations to oldannotations
        # set starting edition for oldannot to either 0 or whatever was last.
        if len(qref.oldannotations) == 0:
            ed = 0
        else:
            ed = qref.oldannotations[-1].edition

        for aref in qref.annotations:
            if aref.edition == 0:  # leave 0th annotation alone.
                continue
            ed += 1
            # make new oldannot using data from aref
            oaref = OldAnnotation.create(
                qgroup=aref.qgroup,
                user=aref.user,
                aimage=aref.aimage,
                edition=ed,
                plom_file=aref.plom_file,
                comment_file=aref.comment_file,
                mark=aref.mark,
                marking_time=aref.marking_time,
                time=aref.time,
                tags=aref.tags,
            )
            # make oapges
            for pref in aref.apages:
                OAPage.create(old_annotation=oaref, order=pref.order, image=pref.image)
            # now delete the apages and then the annotation-image and finally the annotation.
            for pref in aref.apages:
                pref.delete_instance()
            # delete the annotated image from table.
            aref.aimage.delete_instance()
            # finally delete the annotation itself.
            aref.delete_instance()
    log.info("Reverting tq {}.{}".format(test_number, question))
    return [True]
